# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division


from nose.tools import eq_
import logging


from vcfnp.table import VariantsTable
import vcfnp.eff


logger = logging.getLogger(__name__)
debug = logger.debug


def test_tabulate_variants():
    vcf_fn = 'fixture/sample.vcf'

    fields = ('CHROM', 'POS', 'REF', 'ALT', 'FILTER', 'AF')
    tbl = list(VariantsTable(vcf_fn, fields=fields))
    debug(tbl)
    eq_(fields, tbl[0])
    eq_(('19', 111, 'A', 'C', '.', '.'), tbl[1])
    eq_(('20', 17330, 'T', 'A', 'q10', '0.017'), tbl[4])
    eq_(('20', 1110696, 'A', 'G,T', 'PASS', '0.333,0.667'), tbl[5])


def test_tabulate_variants_flatten_filter():
    vcf_fn = 'fixture/sample.vcf'

    fields = ('CHROM', 'POS', 'REF', 'ALT', 'FILTER')
    tbl = list(VariantsTable(vcf_fn, fields=fields, flatten_filter=True))
    expect_fields = ('CHROM', 'POS', 'REF', 'ALT', 'FILTER_q10', 'FILTER_s50',
                     'FILTER_PASS')
    eq_(expect_fields, tbl[0])
    eq_(('19', 111, 'A', 'C', False, False, False), tbl[1])
    eq_(('20', 17330, 'T', 'A', True, False, False), tbl[4])
    eq_(('20', 1110696, 'A', 'G,T', False, False, True), tbl[5])


def test_tabulate_variants_explicit_arity():
    vcf_fn = 'fixture/sample.vcf'

    fields = ('CHROM', 'POS', 'REF', 'ALT', 'AF')
    tbl = list(VariantsTable(vcf_fn, fields=fields,
                             arities={'ALT': 2, 'AF': 2}))
    expect_fields = ('CHROM', 'POS', 'REF', 'ALT_1', 'ALT_2', 'AF_1', 'AF_2')
    eq_(expect_fields, tbl[0])
    eq_(('19', 111, 'A', 'C', '.', '.', '.'), tbl[1])
    eq_(('20', 17330, 'T', 'A', '.', '0.017', '.'), tbl[4])
    eq_(('20', 1110696, 'A', 'G', 'T', '0.333', '0.667'), tbl[5])


def test_tabulate_variants_fill():
    vcf_fn = 'fixture/sample.vcf'

    fields = ('CHROM', 'POS', 'REF', 'ALT', 'FILTER', 'AF')
    tbl = list(VariantsTable(vcf_fn, fields=fields, fill=None))
    debug(tbl)
    eq_(fields, tbl[0])
    eq_(('19', 111, 'A', 'C', None, None), tbl[1])
    eq_(('20', 17330, 'T', 'A', 'q10', '0.017'), tbl[4])
    eq_(('20', 1110696, 'A', 'G,T', 'PASS', '0.333,0.667'), tbl[5])


def test_tabulate_variants_flatten_eff():
    vcf_fn = 'fixture/test12.vcf'

    # test without flattening
    fields = ('CHROM', 'POS', 'REF', 'ALT', 'EFF')
    tbl = list(VariantsTable(vcf_fn, fields=fields, flatteners={'EFF': None}))
    debug(tbl)
    eq_(fields, tbl[0])
    eq_(('1', 889455, 'G', 'A',
         'STOP_GAINED(HIGH|NONSENSE|Cag/Tag|Q236*|749|NOC2L||CODING|'
         'NM_015658|)'), tbl[1])
    eq_(('1', 897062, 'C', 'T',
         'NON_SYNONYMOUS_CODING(MODERATE|MISSENSE|gTt/gGt|V155G||PF3D7_0108900'
         '|||rna_PF3D7_0108900-1|1|1|WARNING_TRANSCRIPT_MULTIPLE_STOP_CODONS'
         ')'), tbl[2])
    eq_(('1', 897063, 'C', 'T', '.'), tbl[3])

    # test with explicit flattening
    fields = ('CHROM', 'POS', 'REF', 'ALT', 'EFF')
    tbl = list(
        VariantsTable(
            vcf_fn,
            fields=fields,
            flatteners={'EFF': (vcfnp.eff.EFF_FIELDS,
                                vcfnp.eff.flatten_eff('NA'))}
        )
    )
    debug(tbl)
    eq_(fields[:4] + ('Effect', 'Effect_Impact', 'Functional_Class',
                      'Codon_Change', 'Amino_Acid_Change', 'Amino_Acid_Length',
                      'Gene_Name', 'Transcript_BioType', 'Gene_Coding',
                      'Transcript_ID', 'Exon'),
        tbl[0])
    eq_(('1', 889455, 'G', 'A', 'STOP_GAINED', 'HIGH', 'NONSENSE',
         'Cag/Tag', 'Q236*', '749', 'NOC2L', 'NA', 'CODING', 'NM_015658',
         'NA'), tbl[1])
    eq_(('1', 897062, 'C', 'T', 'NON_SYNONYMOUS_CODING', 'MODERATE',
         'MISSENSE', 'gTt/gGt', 'V155G', 'NA', 'PF3D7_0108900', 'NA',
         'NA', 'rna_PF3D7_0108900-1', '1'), tbl[2])
    eq_(('1', 897063, 'C', 'T', 'NA', 'NA', 'NA', 'NA', 'NA', 'NA',
         'NA', 'NA', 'NA', 'NA', 'NA'), tbl[3])

    # test with default flattening
    fields = ('CHROM', 'POS', 'REF', 'ALT', 'EFF')
    tbl = list(VariantsTable(vcf_fn, fields=fields, fill=None))
    debug(tbl)
    eq_(fields[:4] + ('Effect', 'Effect_Impact', 'Functional_Class',
                      'Codon_Change', 'Amino_Acid_Change', 'Amino_Acid_Length',
                      'Gene_Name', 'Transcript_BioType', 'Gene_Coding',
                      'Transcript_ID', 'Exon'),
        tbl[0])
    eq_(('1', 889455, 'G', 'A', 'STOP_GAINED', 'HIGH', 'NONSENSE',
         'Cag/Tag', 'Q236*', '749', 'NOC2L', None, 'CODING', 'NM_015658',
         None), tbl[1])
    eq_(('1', 897062, 'C', 'T', 'NON_SYNONYMOUS_CODING', 'MODERATE',
         'MISSENSE', 'gTt/gGt', 'V155G', None, 'PF3D7_0108900', None, None,
         'rna_PF3D7_0108900-1', '1'), tbl[2])
    eq_(('1', 897063, 'C', 'T', None, None, None, None, None, None, None,
         None, None, None, None), tbl[3])


def test_tabulate_variants_flatten_ann():
    vcf_fn = 'fixture/test_ann.vcf'

    # test without flattening
    fields = ('CHROM', 'POS', 'REF', 'ALT', 'ANN')
    tbl = list(VariantsTable(vcf_fn, fields=fields, flatteners={'ANN': None}))
    debug(tbl)
    eq_(fields, tbl[0])
    eq_(('2L', 103, 'C', 'T',
         'T|intergenic_region|MODIFIER|AGAP004677|AGAP004677|intergenic_region'
         '|AGAP004677|||||||||'),
        tbl[1])
    eq_(('2L', 192, 'G', 'A', '.'), tbl[2])
    eq_(('2L', 13513722, 'A', 'T',
         'T|missense_variant|MODERATE|AGAP005273|AGAP005273|transcript|AGAP005'
         '273-RA|VectorBase|1/4|n.17A>T|p.Asp6Val|17/4788|17/-1|6/-1||'),
        tbl[3])

    # test with explicit flattening
    fields = ('CHROM', 'POS', 'REF', 'ALT', 'ANN')
    tbl = list(
        VariantsTable(
            vcf_fn,
            fields=fields,
            flatteners={'ANN': (vcfnp.eff.ANN_FIELDS,
                                vcfnp.eff.flatten_ann('NA'))}
        )
    )
    debug(tbl)
    eq_(fields[:4] + ('Allele', 'Annotation', 'Annotation_Impact', 'Gene_Name',
                      'Gene_ID', 'Feature_Type', 'Feature_ID',
                      'Transcript_BioType', 'Rank', 'HGVS_c', 'HGVS_p',
                      'cDNA_pos', 'cDNA_length', 'CDS_pos', 'CDS_length',
                      'AA_pos', 'AA_length', 'Distance'),
        tbl[0])
    eq_(('2L', 103, 'C', 'T', 'T', 'intergenic_region', 'MODIFIER',
         'AGAP004677', 'AGAP004677', 'intergenic_region', 'AGAP004677', 'NA',
         'NA', 'NA', 'NA', 'NA', 'NA', 'NA', 'NA', 'NA', 'NA', 'NA'),
        tbl[1])
    eq_(('2L', 192, 'G', 'A', 'NA', 'NA', 'NA', 'NA', 'NA', 'NA', 'NA', 'NA',
         'NA', 'NA', 'NA', 'NA', 'NA', 'NA', 'NA', 'NA', 'NA', 'NA'), tbl[2])
    eq_(('2L', 13513722, 'A', 'T', 'T', 'missense_variant', 'MODERATE', 
         'AGAP005273', 'AGAP005273', 'transcript', 'AGAP005273-RA', 
         'VectorBase', '1/4', 'n.17A>T', 'p.Asp6Val', '17', '4788', '17', '-1',
         '6', '-1', 'NA'), tbl[3])

    # test with default flattening
    fields = ('CHROM', 'POS', 'REF', 'ALT', 'ANN')
    tbl = list(VariantsTable(vcf_fn, fields=fields, fill=None))
    debug(tbl)
    eq_(fields[:4] + ('Allele', 'Annotation', 'Annotation_Impact', 'Gene_Name',
                      'Gene_ID', 'Feature_Type', 'Feature_ID',
                      'Transcript_BioType', 'Rank', 'HGVS_c', 'HGVS_p',
                      'cDNA_pos', 'cDNA_length', 'CDS_pos', 'CDS_length',
                      'AA_pos', 'AA_length', 'Distance'),
        tbl[0])
    eq_(('2L', 103, 'C', 'T', 'T', 'intergenic_region', 'MODIFIER',
         'AGAP004677', 'AGAP004677', 'intergenic_region', 'AGAP004677', None,
         None, None, None, None, None, None, None, None, None, None),
        tbl[1])
    eq_(('2L', 192, 'G', 'A', None, None, None, None, None, None, None, None,
         None, None, None, None, None, None, None, None, None, None), tbl[2])
    eq_(('2L', 13513722, 'A', 'T', 'T', 'missense_variant', 'MODERATE', 
         'AGAP005273', 'AGAP005273', 'transcript', 'AGAP005273-RA', 
         'VectorBase', '1/4', 'n.17A>T', 'p.Asp6Val', '17', '4788', '17', '-1',
         '6', '-1', None), tbl[3])



# cython: profile = False
# cython: boundscheck = True
# cython: wraparound = False
# cython: embedsignature = True
from __future__ import print_function, absolute_import, division


from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.map cimport map
from libc.stdlib cimport atoi, atol, atof
from cython.operator cimport dereference as deref
import numpy as np
cimport numpy as np
import logging


from vcfnp.compat cimport b as _b, s as _s
from vcfnp.vcflib cimport VariantCallFile, Variant, FIELD_STRING, \
    FIELD_INTEGER, FIELD_BOOL, FIELD_FLOAT, FIELD_UNKNOWN, ALLELE_NUMBER, \
    GENOTYPE_NUMBER


logger = logging.getLogger(__name__)
debug = logger.debug


cdef size_t npos = -1


cdef extern from "split.h":
    # split a string on a single delimiter character (delim)
    vector[string]& split(const string &s, char delim, vector[string] &elems)
    vector[string]  split(const string &s, char delim)
    # split a string on any character found in the string of delimiters (delims)
    vector[string]& split(const string &s, const string& delims, vector[string] &elems)
    vector[string]  split(const string &s, const string& delims)


cdef char SEMICOLON = b';'
cdef string DOT = b'.'
cdef string GT_DELIMS = b'/|'
cdef string FIELD_NAME_CHROM = b'CHROM'
cdef string FIELD_NAME_POS = b'POS'
cdef string FIELD_NAME_ID = b'ID'
cdef string FIELD_NAME_REF = b'REF'
cdef string FIELD_NAME_ALT = b'ALT'
cdef string FIELD_NAME_QUAL = b'QUAL'
cdef string FIELD_NAME_FILTER = b'FILTER'
cdef string FIELD_NAME_INFO = b'INFO'
cdef string FIELD_NAME_NUM_ALLELES = b'num_alleles'
cdef string FIELD_NAME_IS_SNP = b'is_snp'
cdef string FIELD_NAME_SVLEN = b'svlen'
cdef string FIELD_NAME_IS_CALLED = b'is_called'
cdef string FIELD_NAME_IS_PHASED = b'is_phased'
cdef string FIELD_NAME_GENOTYPE = b'genotype'
cdef string FIELD_NAME_GT = b'GT'


def itervariants(vcf_fns, region, fields, arities, fills, info_types,
                 transformers, filter_ids, flatten_filter, parse_info,
                 condition):
    """Iterate over variants from a VCF file, and generate a tuple for each
    variant suitable for loading into a numpy array."""

    # force to bytes
    vcf_fns = _b(tuple(vcf_fns))
    region = _b(region)
    fields = _b(tuple(fields))
    arities = tuple(arities)
    fills = tuple(fills)
    info_types = tuple(info_types)
    transformers = tuple(transformers)
    filter_ids = _b(tuple(filter_ids))

    # zip up field information for convenience
    fieldspec = tuple(zip(fields, arities, fills, info_types, transformers))

    if condition is None:
        return _itervariants(vcf_fns=vcf_fns, region=region,
                             fieldspec=fieldspec, filter_ids=filter_ids,
                             flatten_filter=flatten_filter,
                             parse_info=parse_info)
    else:
        return _itervariants_with_condition(vcf_fns=vcf_fns, region=region,
                                            fieldspec=fieldspec,
                                            filter_ids=filter_ids,
                                            flatten_filter=flatten_filter,
                                            parse_info=parse_info,
                                            condition=condition)


def _itervariants(tuple vcf_fns, bytes region, tuple fieldspec,
                  tuple filter_ids, bint flatten_filter, bint parse_info):

    # statically typed variables
    cdef VariantCallFile *variant_file
    cdef Variant *variant

    # work through multiple VCFs if provided
    for vcf_fn in vcf_fns:
        variant_file = new VariantCallFile()
        variant_file.open(vcf_fn)
        # set whether INFO field needs to be parsed
        variant_file.parseInfo = parse_info
        # set whether samples fields need to be parsed
        variant_file.parseSamples = False
        if region is not None:
            # set genome region to extract variants from
            region_set = variant_file.setRegion(region)
            if not region_set:
                raise StopIteration
        variant = new Variant(deref(variant_file))

        # iterate over variants
        while _get_next_variant(variant_file, variant):
            yield _mkvrow(variant, fieldspec, filter_ids, flatten_filter)

        # clean up
        del variant_file
        del variant


def _itervariants_with_condition(tuple vcf_fns, bytes region, tuple fieldspec,
                                 tuple filter_ids, bint flatten_filter,
                                 bint parse_info, np.uint8_t[:] condition):

    # statically typed variables
    cdef VariantCallFile *variant_file
    cdef Variant *variant
    cdef long i = 0
    cdef long n = len(condition)

    for vcf_fn in vcf_fns:
        variant_file = new VariantCallFile()
        variant_file.open(vcf_fn)
        variant_file.parseInfo = parse_info
        variant_file.parseSamples = False
        if region is not None:
            region_set = variant_file.setRegion(region)
            if not region_set:
                raise StopIteration
        variant = new Variant(deref(variant_file))

        while i < n and _get_next_variant(variant_file, variant):
            if condition[i]:
                yield _mkvrow(variant, fieldspec, filter_ids, flatten_filter)
            i += 1

        del variant_file
        del variant


cdef inline bool _get_next_variant(VariantCallFile *variant_file,
                                   Variant *variant):
    # break this out into a separate function so we can profile it
    return variant_file.getNextVariant(deref(variant))


cdef inline object _mkvrow(Variant *variant, tuple fieldspec, tuple filter_ids,
                           bint flatten_filter):
    """Make a row of variant data."""
    out = list()
    for f, arity, fill, vcf_type, transformer in fieldspec:
        val = _mkvval(variant, f, arity, fill, vcf_type, transformer,
                      filter_ids)
        if (f == b'FILTER') and flatten_filter:
            out.extend(val)
        else:
            out.append(val)
    return tuple(out)


cdef inline object _mkvval(Variant *variant, string field, int arity,
                           object fill, int vcf_type, transformer,
                           tuple filter_ids):
    if field == FIELD_NAME_CHROM:
        out = variant.sequenceName
    elif field == FIELD_NAME_POS:
        out = variant.position
    elif field == FIELD_NAME_ID:
        out = variant.id
    elif field == FIELD_NAME_REF:
        out = variant.ref
    elif field == FIELD_NAME_ALT:
        out = _mkaltval(variant, arity, fill)
    elif field == FIELD_NAME_QUAL:
        out = variant.quality
    elif field == FIELD_NAME_FILTER:
        out = _mkfilterval(variant, filter_ids)
    elif field == FIELD_NAME_NUM_ALLELES:
        out = <int>(variant.alt.size() + 1)
    elif field == FIELD_NAME_IS_SNP:
        out = _is_snp(variant)
    elif field == FIELD_NAME_SVLEN:
        out = _svlen(variant, arity, fill)
    elif transformer is not None:
        out = transformer(variant.info[field])
    elif vcf_type == FIELD_BOOL:
        # ignore arity, this is a flag
        out = (variant.infoFlags.count(field) > 0)
    else:
        out = _mkval(variant.info[field], arity, fill, vcf_type)
    return out


cdef inline object _mkaltval(Variant *variant, int arity, object fill):
    if arity == 1:
        if variant.alt.size() == 0:
            out = fill
        else:
            out = variant.alt.at(0)
    elif variant.alt.size() == arity:
        out = variant.alt
        out = tuple(out)
    elif variant.alt.size() > arity:
        out = variant.alt
        out = tuple(out[:arity])
    else:
        out = variant.alt
        out += [fill] * (arity-variant.alt.size())
        out = tuple(out)
    return out


cdef inline object _mkfilterval(Variant *variant, tuple filter_ids):
    filters = <list>split(variant.filter, SEMICOLON)
    out = [(f in filters) for f in filter_ids]
    out = tuple(out)
    return out


cdef inline object _is_snp(Variant *variant):
    cdef int i
    cdef bytes alt
    if variant.ref.size() > 1:
        return False
    for i in range(variant.alt.size()):
        alt = variant.alt.at(i)
        if alt not in {b'A', b'C', b'G', b'T'}:
            return False
    return True


cdef inline object _svlen(Variant *variant, int arity, object fill):
    if arity == 1:
        return _svlen_single(variant.ref, variant.alt, fill)
    else:
        return _svlen_multi(variant.ref, variant.alt, arity, fill)


cdef inline object _svlen_single(string ref, vector[string]& alt, object fill):
    if alt.size() > 0:
        return <int>(alt.at(0).size() - ref.size())
    return fill


cdef inline object _svlen_multi(string ref, vector[string]& alt, int arity,
                                object fill):
    cdef int i
    out = list()
    for i in range(arity):
        if i < alt.size():
            out.append(<int>(alt.at(i).size() - ref.size()))
        else:
            out.append(fill)
    return out


cdef inline object _mkval(vector[string]& string_vals, int arity, object fill,
                          int vcf_type):
    if vcf_type == FIELD_FLOAT:
        out = _mkval_double(string_vals, arity, fill)
    elif vcf_type == FIELD_INTEGER:
        out = _mkval_long(string_vals, arity, fill)
    else:
        # make strings by default
        out = _mkval_string(string_vals, arity, fill)
    return out


cdef inline object _mkval_string(vector[string]& string_vals, int arity,
                                 object fill):
    if arity == 1:
        if string_vals.size() > 0:
            return string_vals.at(0)
        else:
            return fill
    else:
        return _mkval_string_multi(string_vals, arity, fill)


cdef inline object _mkval_string_multi(vector[string]& string_vals, int arity,
                                       object fill):
    cdef int i
    out = list()
    for i in range(arity):
        if i < string_vals.size():
            out.append(string_vals.at(i))
        else:
            out.append(fill)
    return out


cdef inline object _mkval_double(vector[string]& string_vals, int arity,
                                 object fill):
    if arity == 1:
        out = _mkval_double_single(string_vals, fill)
    else:
        out = _mkval_double_multi(string_vals, arity, fill)
    return out


cdef inline object _mkval_double_single(vector[string]& string_vals,
                                        object fill):
    cdef double v
    if string_vals.size() > 0:
        return atof(string_vals.at(0).c_str())
    return fill


cdef inline object _mkval_double_multi(vector[string]& string_vals, int arity,
                                       object fill):
    cdef int i
    out = list()
    for i in range(arity):
        if i < string_vals.size():
            out.append(atof(string_vals.at(i).c_str()))
        else:
            out.append(fill)
    return out


cdef inline object _mkval_long(vector[string]& string_vals, int arity,
                               object fill):
    if arity == 1:
        out = _mkval_long_single(string_vals, fill)
    else:
        out = _mkval_long_multi(string_vals, arity, fill)
    return out


cdef inline object _mkval_long_single(vector[string]& string_vals, object fill):
    if string_vals.size() > 0:
        return atol(string_vals.at(0).c_str())
    return fill


cdef inline object _mkval_long_multi(vector[string]& string_vals, int arity,
                                     object fill):
    cdef int i
    out = list()
    for i in range(arity):
        if i < string_vals.size():
            out.append(atol(string_vals.at(i).c_str()))
        else:
            out.append(fill)
    return out


def itercalldata(vcf_fns, region, samples, ploidy, fields, arities, fills,
                 format_types, condition):
    """Iterate over call data (genotypes, etc.) returning tuples suitable for
    loading into a numpy structured array."""

    # force bytes
    vcf_fns = _b(tuple(vcf_fns))
    region = _b(region)
    samples = _b(tuple(samples))
    fields = _b(tuple(fields))
    arities = tuple(arities)
    fills = tuple(fills)
    format_types = tuple(format_types)

    # zip up field parameters
    fieldspec = tuple(zip(fields, arities, fills, format_types))

    if condition is None:
        return _itercalldata(vcf_fns, region, samples, ploidy, fieldspec)
    else:
        return _itercalldata_with_condition(vcf_fns, region, samples, ploidy,
                                            fieldspec, condition)



def _itercalldata(tuple vcf_fns, bytes region, tuple samples, int ploidy,
                  tuple fieldspec):
    cdef VariantCallFile *variant_file
    cdef Variant *variant

    for vcf_fn in vcf_fns:
        variant_file = new VariantCallFile()
        variant_file.open(vcf_fn)
        variant_file.parseInfo = False
        variant_file.parseSamples = True
        if region is not None:
            region_set = variant_file.setRegion(region)
            if not region_set:
                raise StopIteration
        variant = new Variant(deref(variant_file))

        while _get_next_variant(variant_file, variant):
            yield _mkcrow(variant, samples, ploidy, fieldspec)

        del variant_file
        del variant


def _itercalldata_with_condition(tuple vcf_fns, bytes region, tuple samples,
                                 int ploidy, tuple fieldspec,
                                 np.uint8_t[:] condition):
    cdef VariantCallFile *variant_file
    cdef Variant *variant
    cdef long i = 0
    cdef long n = len(condition)

    for vcf_fn in vcf_fns:
        variant_file = new VariantCallFile()
        variant_file.open(vcf_fn)
        variant_file.parseInfo = False
        variant_file.parseSamples = False
        if region is not None:
            region_set = variant_file.setRegion(region)
            if not region_set:
                raise StopIteration
        variant = new Variant(deref(variant_file))

        while i < n:
            # only worth parsing samples if we know we want the variant
            if condition[i]:
                variant_file.parseSamples = True
                if not _get_next_variant(variant_file, variant):
                    break
                yield _mkcrow(variant, samples, ploidy, fieldspec)
            else:
                variant_file.parseSamples = False
                if not _get_next_variant(variant_file, variant):
                    break
            i += 1

        del variant_file
        del variant


cdef inline object _mkcrow(Variant *variant, tuple samples, int ploidy,
                           tuple fieldspec):
    out = [_mkcvals(variant, s, ploidy, fieldspec) for s in samples]
    return tuple(out)


cdef inline object _mkcvals(Variant *variant, string sample, int ploidy,
                            tuple fieldspec):
    out = [_mkcval(variant.samples[sample], ploidy, f, arity, fill, format_type)
           for (f, arity, fill, format_type) in fieldspec]
    return tuple(out)


cdef inline object _mkcval(map[string, vector[string]]& sample_data, int ploidy,
                           string field, int arity, object fill,
                           int format_type):
    if field == FIELD_NAME_IS_CALLED:
        return _is_called(sample_data)
    elif field == FIELD_NAME_IS_PHASED:
        return _is_phased(sample_data)
    elif field == FIELD_NAME_GENOTYPE:
        return _genotype(sample_data, ploidy)
    else:
        return _mkval(sample_data[field], arity, fill, format_type)


cdef inline bool _is_called(map[string, vector[string]]& sample_data):
    cdef vector[string] *gts
    gts = &sample_data[FIELD_NAME_GT]
    if gts.size() == 0:
        return False
    else:
        return gts.at(0).find(b'.') == npos


cdef inline bool _is_phased(map[string, vector[string]]& sample_data):
    cdef vector[string] *gts
    gts = &sample_data[FIELD_NAME_GT]
    if gts.size() == 0:
        return False
    else:
        return gts.at(0).find(b'|') != npos


cdef inline object _genotype(map[string, vector[string]]& sample_data,
                             int ploidy):
    cdef vector[string] *gts
    cdef vector[int] alleles
    cdef vector[string] allele_strings
    cdef int i
    cdef int allele
    gts = &sample_data[FIELD_NAME_GT]
    if gts.size() == 0:
        if ploidy == 1:
            return -1
        else:
            return (-1,) * ploidy
    else:
        split(gts.at(0), GT_DELIMS, allele_strings)
        if ploidy == 1:
            if allele_strings.size() > 0:
                s = allele_strings.at(0)
                if s == b'.':
                    return -1
                else:
                    return atoi(allele_strings.at(0).c_str())
            else:
                return -1
        else:
            for i in range(ploidy):
                if i < allele_strings.size():
                    s = allele_strings.at(i)
                    if s == b'.':
                        alleles.push_back(-1)
                    else:
                        alleles.push_back(atoi(allele_strings.at(i).c_str()))
                else:
                    alleles.push_back(-1)
            return tuple(alleles)


def itervariantstable(vcf_fns, region, fields, arities, info_types, parse_info,
                      filter_ids, flatten_filter, fill, flatteners):

    # force bytes
    vcf_fns = _b(tuple(vcf_fns))
    region = _b(region)
    fields = _b(tuple(fields))
    arities = tuple(arities)
    info_types = tuple(info_types)
    filter_ids = _b(tuple(filter_ids))
    fill = _b(fill)
    flatteners = tuple(flatteners)
    debug(flatteners)

    cdef VariantCallFile *variant_file
    cdef Variant *variant

    for vcf_fn in vcf_fns:
        variant_file = new VariantCallFile()
        variant_file.open(vcf_fn)
        variant_file.parseInfo = parse_info
        variant_file.parseSamples = False
        if region is not None:
            region_set = variant_file.setRegion(region)
            if not region_set:
                raise StopIteration
        variant = new Variant(deref(variant_file))

        while _get_next_variant(variant_file, variant):
            yield _mkvtblrow(variant, fields, arities, info_types, filter_ids,
                             flatten_filter, fill, flatteners)

        del variant_file
        del variant


cdef inline object _mkvtblrow(Variant *variant, tuple fields, tuple arities,
                              tuple info_types,
                              tuple filter_ids, bint flatten_filter,
                              object fill, tuple flatteners):
    out = list()
    cdef string field
    cdef string flt
    for field, arity, vcf_type, flattener in zip(fields, arities, info_types,
                                                 flatteners):
        if field == FIELD_NAME_CHROM:
            out.append(variant.sequenceName)
        elif field == FIELD_NAME_POS:
            out.append(variant.position)
        elif field == FIELD_NAME_ID:
            out.append(variant.id)
        elif field == FIELD_NAME_REF:
            out.append(variant.ref)
        elif field == FIELD_NAME_ALT:
            if arity is not None:
                vals = _mktblval_multi(variant.alt, arity, fill)
                out.extend(vals)
            elif variant.alt.size() == 0:
                out.append(fill)
            else:
                val = b','.join(variant.alt)
                out.append(val)
        elif field == FIELD_NAME_QUAL:
            out.append(variant.quality)
        elif field == FIELD_NAME_FILTER:
            flt = variant.filter
            if flatten_filter:
                out.extend(_mkfilterval(variant, filter_ids))
            elif flt == b'.':
                out.append(fill)
            else:
                out.append(flt)
        elif field == FIELD_NAME_NUM_ALLELES:
            out.append(variant.alt.size() + 1)
        elif field == FIELD_NAME_IS_SNP:
            out.append(_is_snp(variant))
        else:
            if vcf_type == FIELD_BOOL:
                # ignore arity, this is a flag
                val = (variant.infoFlags.count(field) > 0)
                out.append(val)
            else:
                if arity is not None:
                    vals = _mktblval_multi(variant.info[field], arity, fill)
                    out.extend(vals)
                elif flattener is not None:
                    _, t = flattener
                    vals = t(variant.info[field])
                    out.extend(vals)
                elif variant.info[field].size() == 0:
                    out.append(fill)
                else:
                    out.append(b','.join(variant.info[field]))
    # force back to str
    return _s(tuple(out))


cdef inline object _mktblval_multi(vector[string]& string_vals, int arity,
                                   object fill):
    cdef int i
    out = list()
    for i in range(arity):
        if i < string_vals.size():
            out.append(string_vals.at(i))
        else:
            out.append(fill)
    return out



#!/usr/bin/python

import sys
import re
import sqlite3

if len(sys.argv) < 2:
    print "usage", sys.argv[0], " [dbname]"
    print "reads VCF on stdin, and writes output to a sqlite3 db [dbname]"
    exit(1)

dbname = sys.argv[1]

# parse the header
# into a mapping from tag -> type

infotypes = {}
infonumbers = {}

for line in sys.stdin:
    if line.startswith('##INFO'):
        #<ID=XRS,Number=1,Type=Float,
        i = re.search("ID=(.*?),", line)
        n = re.search("Number=(.*?),", line)
        t = re.search("Type=(.*?),", line)
        if i and n and t:
            id = i.groups()[0]
            number = n.groups()[0]
            if number == "A":
                number = -1
            elif number == "G" or int(number) > 1:
                # unclear how to deal with these
                continue
            else:
                number = int(number)
            typestr = t.groups()[0]
            infotypes[id] = typestr
            infonumbers[id] = number
        else:
            continue
    elif line.startswith('##'):
        continue
    else:
        break # header line, sample names etc.

# write the table schema

infotype_to_sqltype = {}
infotype_to_sqltype["Flag"] = "boolean"
infotype_to_sqltype["Integer"] = "integer"
infotype_to_sqltype["Float"] = "real"
infotype_to_sqltype["String"] = "text"

tablecmd = """create table alleles"""
specs = ["CHROM text",
        "POS integer",
        "ID text",
        "REF text",
        "ALT text",
        "QUAL real",
        "FILTER text"]

sorted_fields = sorted(infotypes.keys())
for field in sorted_fields:
    infotype = infotypes[field]
    sqltype = infotype_to_sqltype[infotype]
    field = field.replace(".", "_") # escape periods, which are not allowed
    specs.append(field + " " + sqltype)

tablecmd += " (" + ", ".join(specs) + ")"

conn = sqlite3.connect(dbname)
conn.execute(tablecmd)

# for each record
# parse the record
# for each allele

for line in sys.stdin:
    fields = line.split('\t')
    chrom, pos, id, ref, alts, qual, filter, info = fields[:8]
    alts = alts.split(",")
    altindex = 0
    chrom = "\'" + chrom + "\'"
    id = "\'" + id + "\'"
    ref = "\'" + ref + "\'"
    filter = "\'" + filter + "\'"
    for alt in alts:
        alt = "\'" + alt + "\'"
        info_values = {}
        for pair in info.split(";"):
            if pair.find("=") is not -1:
                pair = pair.split("=")
                key = pair[0]
                value = pair[1]
                if not infonumbers.has_key(key):
                    continue
                if infonumbers[key] == -1:
                    values = value.split(",")
                    value = values[altindex]
                info_values[key] = value
            else:
               # boolean flag
                info_values[pair] = "1"
        ordered_insertion = []
        for field in sorted_fields:
            value = "null"
            if info_values.has_key(field):
                value = info_values[field]
                if infotypes[field] == "String":
                    value = "\'" + value + "\'"
            else:
                # missing flag means "false" for that flag
                if infotypes[field] == "Flag":
                    value = "0"
            ordered_insertion.append(value)
        cmd = "insert into alleles values (" \
            + ", ".join([chrom, pos, id, ref, alt, qual, filter]) \
            + ", " \
            + ", ".join(ordered_insertion) + ")"
        conn.execute(cmd)
        altindex += 1

conn.commit()

# TODO ignoring samples (for now)

# add indexes everywhere?

conn.close()



#!/usr/bin/python
# -*- coding: utf-8 -*-
# multipermute.py - permutations of a multiset
# Erik Garrison <erik.garrison@bc.edu> 2010

"""
This module encodes functions to generate the permutations of a multiset
following this algorithm:

Algorithm 1 
Visits the permutations of multiset E. The permutations are stored
in a singly-linked list pointed to by head pointer h. Each node in the linked
list has a value field v and a next field n. The init(E) call creates a
singly-linked list storing the elements of E in non-increasing order with h, i,
and j pointing to its first, second-last, and last nodes, respectively. The
null pointer is given by φ. Note: If E is empty, then init(E) should exit.
Also, if E contains only one element, then init(E) does not need to provide a
value for i.

[h, i, j] ← init(E) 
visit(h) 
while j.n ≠ φ orj.v <h.v do
    if j.n ≠    φ and i.v ≥ j.n.v then 
        s←j
    else
        s←i 
    end if
    t←s.n 
    s.n ← t.n 
    t.n ← h 
    if t.v < h.v then
        i←t 
    end if
    j←i.n 
    h←t 
    visit(h)
end while

... from "Loopless Generation of Multiset Permutations using a Constant Number
of Variables by Prefix Shifts."  Aaron Williams, 2009
"""

class ListElement:
    def __init__(self, value, next):
        self.value = value
        self.next = next
    def nth(self, n):
        o = self
        i = 0
        while i < n and o.next is not None:
            o = o.next
            i += 1
        return o

def init(multiset):
    multiset.sort() # ensures proper non-increasing order
    h = ListElement(multiset[0], None)
    for item in multiset[1:]:
        h = ListElement(item, h)
    return h, h.nth(len(multiset) - 2), h.nth(len(multiset) - 1)

def visit(h):
    """Converts our bespoke linked list to a python list."""
    o = h
    l = []
    while o is not None:
        l.append(o.value)
        o = o.next
    return l

def permutations(multiset):
    """Generator providing all multiset permutations of a multiset."""
    h, i, j = init(multiset)
    yield visit(h)
    while j.next is not None or j.value < h.value:
        if j.next is not None and i.value >= j.next.value:
            s = j
        else:
            s = i
        t = s.next
        s.next = t.next
        t.next = h
        if t.value < h.value:
            i = t
        j = i.next
        h = t
        yield visit(h)

if __name__ == '__main__':
    import sys
    multiset = sys.argv[1:]
    if multiset != []:
        for permutation in permutations(multiset):
            for item in permutation:
                print item,
            print
    else:
        print "usage", sys.argv[0], "<multiset>"



from __future__ import print_function, division, absolute_import


import os
from nose.tools import eq_, assert_almost_equal
import numpy as np
import logging


import vcfnp
from vcfnp.array import variants, calldata, calldata_2d


logger = logging.getLogger(__name__)
debug = logger.debug


def test_variants():
    a = variants('fixture/sample.vcf', arities={'ALT': 2, 'AC': 2})
    debug(repr(a))
    eq_(9, len(a))

    eq_(b'19', a[0]['CHROM'])
    eq_(111, a[0]['POS'])
    eq_(b'rs6054257', a[2]['ID'])
    eq_(b'A', a[0]['REF'])
    eq_(b'ATG', a[8]['ALT'][1])
    eq_(10.0, a[1]['QUAL'])
    eq_(True, a[2]['FILTER']['PASS'])
    eq_(False, a[3]['FILTER']['PASS'])
    eq_(True, a[3]['FILTER']['q10'])
    eq_(2, a[0]['num_alleles'])
    eq_(False, a[5]['is_snp'])

    # INFO fields
    eq_(3, a[2]['NS'])
    eq_(.5, a[2]['AF'])
    eq_(True, a[2]['DB'])
    eq_((3, 1), tuple(a[6]['AC']))


def test_variants_flatten_filter():
    a = variants('fixture/sample.vcf', flatten_filter=True)
    debug(a)
    debug(a.dtype)
    eq_(True, a[2]['FILTER_PASS'])
    eq_(False, a[3]['FILTER_PASS'])
    eq_(True, a[3]['FILTER_q10'])


def test_variants_region():
    a = variants('fixture/sample.vcf.gz', region='20')
    eq_(6, len(a))
    
    
def test_variants_region_empty():
    a = variants('fixture/sample.vcf.gz', region='18')
    eq_(0, len(a))
    a = variants('fixture/sample.vcf.gz', region='19:113-200')
    eq_(0, len(a))


def test_variants_count():
    a = variants('fixture/sample.vcf', count=3)
    eq_(3, len(a))
    
    
def test_variants_exclude_fields():
    a = variants('fixture/sample.vcf', exclude_fields=['ID', 'FILTER'])
    assert 'CHROM' in a.dtype.names
    assert 'ID' not in a.dtype.names
    assert 'FILTER' not in a.dtype.names
    
    
def test_variants_slice():
    a = variants('fixture/sample.vcf.gz')
    eq_(b'rs6054257', a['ID'][2])
    a = variants('fixture/sample.vcf.gz', slice_args=(0, None, 2))
    eq_(b'rs6054257', a['ID'][1])
    
    
def test_calldata():
    a = calldata('fixture/sample.vcf')
    debug(repr(a))
    eq_(b'0|0', a[0]['NA00001']['GT'])
    eq_(True, a[0]['NA00001']['is_called'])
    eq_(True, a[0]['NA00001']['is_phased'])
    eq_((0, 0), tuple(a[0]['NA00001']['genotype']))
    eq_((-1, -1), tuple(a[6]['NA00003']['genotype']))
    eq_((-1, -1), tuple(a[7]['NA00003']['genotype']))
    eq_((10, 10), tuple(a[0]['NA00001']['HQ']))


def test_calldata_region():
    a = calldata('fixture/sample.vcf.gz', region='20')
    eq_(6, len(a))


def test_calldata_region_empty():
    a = calldata('fixture/sample.vcf.gz', region='18')
    eq_(0, len(a))
    a = calldata('fixture/sample.vcf.gz', region='19:113-200')
    eq_(0, len(a))


def test_condition():
    v = variants('fixture/sample.vcf')
    eq_(9, len(v))
    c = calldata('fixture/sample.vcf', condition=v['FILTER']['PASS'])
    eq_(5, len(c))
    vf = variants('fixture/sample.vcf', condition=v['FILTER']['PASS'])
    eq_(5, len(vf))
    

def test_variable_calldata():
    c = calldata('fixture/test1.vcf')
    eq_((1, 0), tuple(c['test2']['AD'][0]))
    eq_((1, 0), tuple(c['test2']['AD'][1]))
    eq_((1, 0), tuple(c['test2']['AD'][2]))
    eq_(b'.', c['test2']['GT'][0])
    eq_(b'0', c['test2']['GT'][1])
    eq_(b'1', c['test2']['GT'][2])
    
    
def test_missing_calldata():
    c = calldata('fixture/test1.vcf')

    # first variant, second sample
    eq_(b'.', c['test2']['GT'][0])
    eq_((-1, -1), tuple(c['test2']['genotype'][0]))
    eq_((1, 0), tuple(c['test2']['AD'][0]))  # data are present

    # third variant, third sample
    eq_(b'.', c['test3']['GT'][2])
    eq_((-1, -1), tuple(c['test3']['genotype'][2]))
    eq_((0, 0), tuple(c['test3']['AD'][2]))  # default fill

    # third variant, fourth sample
    eq_(b'./.', c['test4']['GT'][2])
    eq_((-1, -1), tuple(c['test4']['genotype'][2]))
    eq_((0, 0), tuple(c['test4']['AD'][2]))  # default fill


def test_missing_calldata_cleared():
    c = calldata('fixture/test32.vcf')['AC0093-C']

    # first variant, non-missing
    eq_(b'0/0', c['GT'][0])
    eq_((0, 0), tuple(c['genotype'][0]))
    eq_(8, c['DP'][0])
    eq_(3, c['GQ'][0])

    # second variant, missing
    eq_(b'./.', c['GT'][1])
    eq_((-1, -1), tuple(c['genotype'][1]))
    eq_(0, c['DP'][1])  # should be default fill value
    eq_(0, c['GQ'][1])  # should be default fill value


def test_override_vcf_types():
    v = variants('fixture/test4.vcf')
    eq_(0, v['MQ0FractionTest'][2])
    v = variants('fixture/test4.vcf', vcf_types={'MQ0FractionTest': 'Float'})
    assert_almost_equal(0.03, v['MQ0FractionTest'][2], )


def test_variants_transformers_eff():

    def _test(v):

        eq_(b'STOP_GAINED', v['EFF']['Effect'][0])
        eq_(b'HIGH', v['EFF']['Effect_Impact'][0])
        eq_(b'NONSENSE', v['EFF']['Functional_Class'][0])
        eq_(b'Cag/Tag', v['EFF']['Codon_Change'][0])
        eq_(b'Q236*', v['EFF']['Amino_Acid_Change'][0])
        eq_(749, v['EFF']['Amino_Acid_Length'][0])
        eq_(b'NOC2L', v['EFF']['Gene_Name'][0])
        eq_(b'.', v['EFF']['Transcript_BioType'][0])
        eq_(1, v['EFF']['Gene_Coding'][0])
        eq_(b'NM_015658', v['EFF']['Transcript_ID'][0])
        eq_(-1, v['EFF']['Exon'][0])

        eq_(b'NON_SYNONYMOUS_CODING', v['EFF']['Effect'][1])
        eq_(b'MODERATE', v['EFF']['Effect_Impact'][1])
        eq_(b'MISSENSE', v['EFF']['Functional_Class'][1])
        eq_(b'gTt/gGt', v['EFF']['Codon_Change'][1])
        eq_(b'V155G', v['EFF']['Amino_Acid_Change'][1])
        eq_(-1, v['EFF']['Amino_Acid_Length'][1])
        eq_(b'PF3D7_0108900', v['EFF']['Gene_Name'][1])
        eq_(b'.', v['EFF']['Transcript_BioType'][1])
        eq_(-1, v['EFF']['Gene_Coding'][1])
        eq_(b'rna_PF3D7_0108900-1', v['EFF']['Transcript_ID'][1])
        eq_(1, v['EFF']['Exon'][1])

        eq_(b'.', v['EFF']['Effect'][2])
        eq_(b'.', v['EFF']['Effect_Impact'][2])
        eq_(b'.', v['EFF']['Functional_Class'][2])
        eq_(b'.', v['EFF']['Codon_Change'][2])
        eq_(b'.', v['EFF']['Amino_Acid_Change'][2])
        eq_(-1, v['EFF']['Amino_Acid_Length'][2])
        eq_(b'.', v['EFF']['Gene_Name'][2])
        eq_(b'.', v['EFF']['Transcript_BioType'][2])
        eq_(-1, v['EFF']['Gene_Coding'][2])
        eq_(b'.', v['EFF']['Transcript_ID'][2])
        eq_(-1, v['EFF']['Exon'][2])

    varr = variants('fixture/test12.vcf',
                    dtypes={'EFF': vcfnp.eff.EFF_DEFAULT_DTYPE},
                    arities={'EFF': 1},
                    transformers={'EFF': vcfnp.eff.eff_default_transformer()})
    _test(varr)

    # test EFF is included in defaults
    varr = variants('fixture/test12.vcf')
    _test(varr)


def test_variants_transformers_ann():

    def _test(v):

        eq_(b'T', v['ANN']['Allele'][0])
        eq_(b'intergenic_region', v['ANN']['Annotation'][0])
        eq_(b'MODIFIER', v['ANN']['Annotation_Impact'][0])
        eq_(b'AGAP004677', v['ANN']['Gene_Name'][0])
        eq_(b'AGAP004677', v['ANN']['Gene_ID'][0])
        eq_(b'intergenic_region', v['ANN']['Feature_Type'][0])
        eq_(b'AGAP004677', v['ANN']['Feature_ID'][0])
        eq_(b'.', v['ANN']['Transcript_BioType'][0])
        eq_(-1, v['ANN']['Rank'][0])
        eq_(b'.', v['ANN']['HGVS_c'][0])
        eq_(b'.', v['ANN']['HGVS_p'][0])
        eq_(-1, v['ANN']['cDNA_pos'][0])
        eq_(-1, v['ANN']['cDNA_length'][0])
        eq_(-1, v['ANN']['CDS_pos'][0])
        eq_(-1, v['ANN']['CDS_length'][0])
        eq_(-1, v['ANN']['AA_pos'][0])
        eq_(-1, v['ANN']['AA_length'][0])
        eq_(-1, v['ANN']['Distance'][0])

        eq_(b'.', v['ANN']['Allele'][1])
        eq_(b'.', v['ANN']['Annotation'][1])
        eq_(b'.', v['ANN']['Annotation_Impact'][1])
        eq_(b'.', v['ANN']['Gene_Name'][1])
        eq_(b'.', v['ANN']['Gene_ID'][1])
        eq_(b'.', v['ANN']['Feature_Type'][1])
        eq_(b'.', v['ANN']['Feature_ID'][1])
        eq_(b'.', v['ANN']['Transcript_BioType'][1])
        eq_(-1, v['ANN']['Rank'][1])
        eq_(b'.', v['ANN']['HGVS_c'][1])
        eq_(b'.', v['ANN']['HGVS_p'][1])
        eq_(-1, v['ANN']['cDNA_pos'][1])
        eq_(-1, v['ANN']['cDNA_length'][1])
        eq_(-1, v['ANN']['CDS_pos'][1])
        eq_(-1, v['ANN']['CDS_length'][1])
        eq_(-1, v['ANN']['AA_pos'][1])
        eq_(-1, v['ANN']['AA_length'][1])
        eq_(-1, v['ANN']['Distance'][1])

        eq_(b'T', v['ANN']['Allele'][2])
        eq_(b'missense_variant', v['ANN']['Annotation'][2])
        eq_(b'MODERATE', v['ANN']['Annotation_Impact'][2])
        eq_(b'AGAP005273', v['ANN']['Gene_Name'][2])
        eq_(b'AGAP005273', v['ANN']['Gene_ID'][2])
        eq_(b'transcript', v['ANN']['Feature_Type'][2])
        eq_(b'AGAP005273-RA', v['ANN']['Feature_ID'][2])
        eq_(b'VectorBase', v['ANN']['Transcript_BioType'][2])
        eq_(1, v['ANN']['Rank'][2])
        eq_(b'n.17A>T', v['ANN']['HGVS_c'][2])
        eq_(b'p.Asp6Val', v['ANN']['HGVS_p'][2])
        eq_(17, v['ANN']['cDNA_pos'][2])
        eq_(4788, v['ANN']['cDNA_length'][2])
        eq_(17, v['ANN']['CDS_pos'][2])
        eq_(-1, v['ANN']['CDS_length'][2])
        eq_(6, v['ANN']['AA_pos'][2])
        eq_(-1, v['ANN']['AA_length'][2])
        eq_(-1, v['ANN']['Distance'][2])

    varr = variants('fixture/test_ann.vcf',
                    dtypes={'ANN': vcfnp.eff.ANN_DEFAULT_DTYPE},
                    arities={'ANN': 1},
                    transformers={'ANN': vcfnp.eff.ann_default_transformer()})
    _test(varr)

    # test ANN is included in defaults
    varr = variants('fixture/test_ann.vcf')
    _test(varr)


def test_svlen():
    # v = variants('fixture/test13.vcf').view(np.recarray)
    # assert hasattr(v, 'svlen')
    # eq_(0, v.svlen[0])
    # eq_(1, v.svlen[1])
    # eq_(-1, v.svlen[2])
    # eq_(3, v.svlen[3])
    # eq_(3, v.svlen[4])
    v = variants('fixture/test13.vcf', arities={'svlen': 2}).view(np.recarray)
    # assert hasattr(v, 'svlen')
    # eq_((3, 0), tuple(v.svlen[3]))
    # eq_((3, -2), tuple(v.svlen[4]))


def test_duplicate_field_definitions():
    variants('fixture/test10.vcf')
    # should not raise, but print useful message to stderr
    calldata('fixture/test10.vcf')
    # should not raise, but print useful message to stderr


def test_missing_info_definition():
    # INFO field DP not declared in VCF header
    v = variants('fixture/test14.vcf', fields=['DP'])
    eq_(b'14', v[2]['DP'])  # default is string
    v = variants('fixture/test14.vcf', fields=['DP'],
                 vcf_types={'DP': 'Integer'})
    eq_(14, v[2]['DP'])
    # what about a field which isn't present at all?
    v = variants('fixture/test14.vcf', fields=['FOO'])
    eq_(b'', v[2]['FOO'])  # default missing value for string field


def test_missing_format_definition():
    # FORMAT field DP not declared in VCF header
    c = calldata('fixture/test14.vcf', fields=['DP'],
                 vcf_types={'DP': 'Integer'})
    eq_(1, c[2]['NA00001']['DP'])


def test_explicit_pass_definition():
    # explicit PASS FILTER definition
    variants('fixture/test16.vcf')
    # should not raise


def test_caching():
    vcf_fn = 'fixture/sample.vcf.gz'

    cache_fn = vcfnp.array._mk_cache_fn(vcf_fn, array_type='variants')
    debug(cache_fn)
    if os.path.exists(cache_fn):
        os.remove(cache_fn)
    a = variants(vcf_fn, cache=True, verbose=True)
    a2 = np.load(cache_fn)
    assert np.all(a == a2)

    cache_fn = vcfnp.array._mk_cache_fn(vcf_fn, array_type='calldata')
    debug(cache_fn)
    if os.path.exists(cache_fn):
        os.remove(cache_fn)
    a = calldata(vcf_fn, cache=True, verbose=True)
    a2 = np.load(cache_fn)
    assert np.all(a == a2)

    cache_fn = vcfnp.array._mk_cache_fn(vcf_fn, array_type='calldata_2d')
    debug(cache_fn)
    if os.path.exists(cache_fn):
        os.remove(cache_fn)
    a = calldata_2d(vcf_fn, cache=True, verbose=True)
    a2 = np.load(cache_fn)
    assert np.all(a == a2)


def test_caching_cachedir():
    vcf_fn = 'fixture/sample.vcf.gz'
    cachedir = 'fixture/custom.vcfnp_cache/foo'

    cache_fn = vcfnp.array._mk_cache_fn(vcf_fn, array_type='variants',
                                        cachedir=cachedir)
    debug(cache_fn)
    if os.path.exists(cache_fn):
        os.remove(cache_fn)
    a = variants(vcf_fn, cache=True, verbose=True, cachedir=cachedir)
    a2 = np.load(cache_fn)
    assert np.all(a == a2)

    cache_fn = vcfnp.array._mk_cache_fn(vcf_fn, array_type='calldata',
                                        cachedir=cachedir)
    debug(cache_fn)
    if os.path.exists(cache_fn):
        os.remove(cache_fn)
    a = calldata(vcf_fn, cache=True, verbose=True, cachedir=cachedir)
    a2 = np.load(cache_fn)
    assert np.all(a == a2)

    cache_fn = vcfnp.array._mk_cache_fn(vcf_fn, array_type='calldata_2d',
                                        cachedir=cachedir)
    debug(cache_fn)
    if os.path.exists(cache_fn):
        os.remove(cache_fn)
    a = calldata_2d(vcf_fn, cache=True, verbose=True, cachedir=cachedir)
    a2 = np.load(cache_fn)
    assert np.all(a == a2)



from libcpp cimport bool
from libcpp.string cimport string
from libcpp.vector cimport vector
from libcpp.map cimport map
from libcpp.pair cimport pair


cdef extern from "split.h":
    # split a string on a single delimiter character (delim)
    vector[string]& split(const string &s, char delim, vector[string] &elems)
    vector[string]  split(const string &s, char delim)
    # split a string on any character found in the string of delimiters (delims)
    vector[string]& split(const string &s, const string& delims, vector[string] &elems)
    vector[string]  split(const string &s, const string& delims)


cdef extern from "Variant.h" namespace "vcf":

    cdef enum VariantFieldType:
        FIELD_FLOAT = 0
        FIELD_INTEGER
        FIELD_BOOL
        FIELD_STRING
        FIELD_UNKNOWN
    
    cdef enum VariantFieldNumber:
        ALLELE_NUMBER = -2
        GENOTYPE_NUMBER = -1
    
    const int INDEX_NONE = -1
    const int NULL_ALLELE = -1
    
    VariantFieldType typeStrToFieldType(string& typeStr)

    cdef cppclass VariantCallFile:
#        istream* file
#        Tabix* tabixFile
        bool usingTabix
        string header
        string line
        string fileformat
        string fileDate
        string source
        string reference
        string phasing
        map[string, VariantFieldType] infoTypes
        map[string, int] infoCounts
        map[string, VariantFieldType] formatTypes
        map[string, int] formatCounts
        vector[string] sampleNames
        bool parseInfo
        bool parseSamples
        bool _done
        void updateSamples(vector[string]& newSampleNames)
        void addHeaderLine(string line)
        void removeInfoHeaderLine(string line)
        void removeGenoHeaderLine(string line)
        vector[string] infoIds()
        vector[string] formatIds()
        vector[string] filterIds()
        bool open(string& filename) 
        bool openFile(string& filename)
        bool openTabix(string& filename)
#        bool open(istream& stream) 
#        bool open(ifstream& stream) 
        bool openForOutput(string& headerStr)
        bool is_open()
        bool eof()
        bool done()
        bool parseHeader(string& headerStr)
        bool parseHeader()
        bool getNextVariant(Variant& var)
        bool setRegion(string region)
        bool setRegion(string seq, long int start, long int end)


    cdef cppclass VariantAllele:
        string ref
        string alt
        string repr
        long position
        VariantAllele(string r, string a, long p)
        
    
    cdef cppclass Variant:
        string sequenceName
        long position
        string id
        string ref
        vector[string] alt
        vector[string] alleles
        map[string, int] altAlleleIndexes
        map[string, vector[VariantAllele] ] parsedAlternates(bool includePreviousBaseForIndels,
                                 bool useMNPs,
                                 bool useEntropy,
                                 float matchScore,
                                 float mismatchScore,
                                 float gapOpenPenalty,
                                 float gapExtendPenalty,
                                 float repeatGapExtendPenalty,
                                 string flankingRefLeft,
                                 string flankingRefRight)
        map[string, string] extendedAlternates(long int newPosition, long int length)
        string originalLine
        string filter
        double quality
        VariantFieldType infoType(string& key)
        map[string, vector[string]] info
        map[string, bool] infoFlags
        VariantFieldType formatType(string& key)
        vector[string] format
        map[string, map[string, vector[string]]] samples
        vector[string] sampleNames
        vector[string] outputSampleNames
        VariantCallFile* vcf
        void removeAlt(string& altallele)
        Variant()
        Variant(VariantCallFile& v)
        void setVariantCallFile(VariantCallFile& v)
        void setVariantCallFile(VariantCallFile* v)
        void parse(string& line, bool parseInfo, bool parseSamples)
        void addFilter(string& tag)
        bool getValueBool(string& key, string& sample, int index)
        double getValueFloat(string& key, string& sample, int index)
        string getValueString(string& key, string& sample, int index)
        bool getSampleValueBool(string& key, string& sample, int index)
        double getSampleValueFloat(string& key, string& sample, int index)
        string getSampleValueString(string& key, string& sample, int index)
        bool getInfoValueBool(string& key, int index)
        double getInfoValueFloat(string& key, int index)
        string getInfoValueString(string& key, int index)
#        void printAlt(ostream& out)
#        void printAlleles(ostream& out)
        int getAltAlleleIndex(string& allele)
        void updateAlleleIndexes()
        void addFormatField(string& key)
        void setOutputSampleNames(vector[string]& outputSamples)
        map[pair[int, int], int] getGenotypeIndexesDiploid()
        int getNumSamples()
        int getNumValidGenotypes()
        

cdef class PyVariantCallFile:

    cdef VariantCallFile *thisptr



from __future__ import print_function, division, absolute_import


from cython.operator cimport dereference as deref
from vcfnp.compat cimport s as _s, b as _b
from collections import namedtuple


# expose these constants to Python
TYPE_INTEGER = FIELD_INTEGER
TYPE_FLOAT = FIELD_FLOAT
TYPE_STRING = FIELD_STRING
TYPE_BOOL = FIELD_BOOL
TYPE_UNKNOWN = FIELD_UNKNOWN
NUMBER_ALLELE = ALLELE_NUMBER
NUMBER_GENOTYPE = GENOTYPE_NUMBER


VariantTuple = namedtuple(
    'Variant',
    ['CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER', 'INFO', 'samples']
)


cdef class PyVariantCallFile:

    def __cinit__(self, filename):
        self.thisptr = new VariantCallFile()
        self.thisptr.open(_b(filename))

    def __init__(self, filename):
        pass

    def __dealloc__(self):
        del self.thisptr        
        
    def __len__(self):
        cdef Variant var
        cdef long n = 0
        var.setVariantCallFile(self.thisptr)
        while self.thisptr.getNextVariant(var):
            n += 1
        return n

    def __iter__(self):
        cdef Variant *var
        cdef vector[string] filters
        cdef char semicolon = b';'
        var = new Variant(deref(self.thisptr))
        while self.thisptr.getNextVariant(deref(var)):
            # split the filter field here in C++ to avoid having to do it in
            # Python later
            filters = split(var.filter, semicolon)
            v = VariantTuple(
                _s(var.sequenceName),
                var.position,
                _s(var.id),
                _s(var.ref),
                _s(var.alt),
                var.quality,
                _s(filters),
                _s(var.info),
                _s(var.samples)
            )
            yield v
        del var
        
    def set_region(self, *args):
        if len(args) == 1:
            region = _b(args[0])
            self.thisptr.setRegion(region)
        elif len(args) == 3:
            chrom, start, stop = args
            chrom = _b(chrom)
            start = int(start)
            stop = int(stop)
            self.thisptr.setRegion(chrom, start, stop)
        else:
            raise Exception('Either provide a single region string '
                            'or provide chrom, start, stop.')

    property info_ids:
        def __get__(self):
            l = _s(<list>self.thisptr.infoIds())
            return l

    property format_ids:
        def __get__(self):
            l = _s(<list>self.thisptr.formatIds())
            return l

    property filter_ids:
        def __get__(self):
            l = _s(<list>self.thisptr.filterIds())
            return l

    property info_types:
        def __get__(self):
            d = _s(<dict>self.thisptr.infoTypes)
            return d

    property format_types:
        def __get__(self):
            d = _s(<dict>self.thisptr.formatTypes)
            return d

    property info_counts:
        def __get__(self):
            d = _s(<dict>self.thisptr.infoCounts)
            return d

    property format_counts:
        def __get__(self):
            d = _s(<dict>self.thisptr.formatCounts)
            return d

    property parse_samples:
        def __get__(self):
            return <bool>self.thisptr.parseSamples
        def __set__(self, v):
            self.thisptr.parseSamples = v

    property header:
        def __get__(self):
            s = _s(self.thisptr.header)
            return s

    property file_format: # [sic] no camel case
        def __get__(self):
            s = _s(self.thisptr.fileformat)
            return s
        
    property file_date:
        def __get__(self):
            s = _s(self.thisptr.fileDate)
            return s
        
    property source:
        def __get__(self):
            s = _s(self.thisptr.source)
            return s
        
    property reference:
        def __get__(self):
            s = _s(self.thisptr.reference)
            return s
        
    property phasing:
        def __get__(self):
            s = _s(self.thisptr.phasing)
            return s
        
    property sample_names:
        def __get__(self):
            l = _s(<list>self.thisptr.sampleNames)
            return l



#!/usr/bin/env python
#
# The MIT License
#
# Copyright (c) 2011 Seoul National University.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Contact: Hyeshik Chang <hyeshik@snu.ac.kr>

from distutils.core import setup, Extension

# Change this to True when you need the knetfile support.
USE_KNETFILE = False

TABIX_SOURCE_FILES = [
    '../bgzf.c', '../bgzip.c', '../index.c', '../knetfile.c', '../kstring.c'
]

define_options = [('_FILE_OFFSET_BITS', 64)]
if USE_KNETFILE:
    define_options.append(('_USE_KNETFILE', 1))

ext_modules = [Extension("tabix", ["tabixmodule.c"] + TABIX_SOURCE_FILES,
                         include_dirs=['..'],
                         libraries=['z'],
                         define_macros=define_options)]

setup (name = 'tabix',
       version = '1.0',
       description = 'Python interface to tabix, a generic indexer '
                     'for TAB-delimited genome position files',
       author = 'Hyeshik Chang',
       author_email = 'hyeshik@snu.ac.kr',
       license = 'MIT',
       ext_modules = ext_modules
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division


# python standard library dependencies
import logging
import inspect
import sys
import os
import time
from datetime import datetime
from itertools import islice


# third party dependencies
import numpy as np


# internal dependencies
from vcfnp.vcflib import PyVariantCallFile, TYPE_STRING, NUMBER_ALLELE, \
    NUMBER_GENOTYPE
from vcfnp.compat import string_types
from vcfnp.iter import itervariants, itercalldata
import vcfnp.config as config


logger = logging.getLogger(__name__)
debug = lambda msg: logger.debug('%s: %s' % (inspect.stack()[1][3], msg))


def variants(vcf_fn, region=None, fields=None, exclude_fields=None, dtypes=None,
             arities=None, fills=None, transformers=None, vcf_types=None,
             count=None, progress=0, logstream=None, condition=None,
             slice_args=None, flatten_filter=False, verbose=True, cache=False,
             cachedir=None, skip_cached=False):
    """
    Load an numpy structured array with data from the fixed fields of a VCF file
    (including INFO).

    Parameters
    ----------

    vcf_fn: string or list
        Name of the VCF file or list of file names.
    region: string, optional
        Region to extract, e.g., 'chr1' or 'chr1:0-100000'.
    fields: list or array-like, optional
        List of fields to extract from the VCF.
    exclude_fields: list or array-like, optional
        Fields to exclude from extraction.
    dtypes: dict or dict-like, optional
        Dictionary cotaining dtypes to use instead of the default inferred ones.
    arities: dict or dict-like, optional
        Dictionary containing field:integer mappings used to override the number
        of values to expect.
    fills: dict or dict-like, optional
        Dictionary containing field:fillvalue mappings used to override the
        defaults used for missing values.
    transformers: dict or dict-like, optional
        Dictionary containing field:function mappings used to preprocess
        any values prior to loading into array.
    vcf_types: dict or dict-like, optional
        Dictionary containing field:string mappings used to override any
        bogus type declarations in the VCF header (e.g., MQ0Fraction declared
        as Integer).
    count: int, optional
        Attempt to extract a specific number of records.
    progress: int, optional
        If greater than 0, log progress.
    logstream: file or file-like object, optional
        Stream to use for logging progress.
    condition: array, optional
        Boolean array defining which rows to load.
    slice_args: tuple or list, optional
        Slice of the underlying iterator, e.g., (0, 1000, 10) takes every
        10th row from the first 1000.
    flatten_filter: bool, optional
        Return FILTER as multiple boolean fields, e.g., FILTER_PASS,
        FILTER_LowQuality, etc.
    verbose: bool, optional
        Log more messages.
    cache: bool, optional
        If True, save the resulting numpy array to disk, and load from the
        cache if present rather than rebuilding from the VCF.
    cachedir: string, optional
        Manually specify the directory to use to store cache files.
    skip_cached: bool, optional
        If True and cache file is fresh, do not load and return None.

    Examples
    --------

        >>> from vcfnp import variants
        >>> v = variants('fixture/sample.vcf')
        >>> v
        array([ ('19', 111, '.', 'A', 'C', 9.600000381469727, (False, False, False), 2, True, 0, '.', 0, 0.0, 0, False, 0, False, 0),
               ('19', 112, '.', 'A', 'G', 10.0, (False, False, False), 2, True, 0, '.', 0, 0.0, 0, False, 0, False, 0),
               ('20', 14370, 'rs6054257', 'G', 'A', 29.0, (False, False, True), 2, True, 0, '.', 0, 0.5, 0, True, 14, True, 3),
               ('20', 17330, '.', 'T', 'A', 3.0, (True, False, False), 2, True, 0, '.', 0, 0.017000000923871994, 0, False, 11, False, 3),
               ('20', 1110696, 'rs6040355', 'A', 'G', 67.0, (False, False, True), 3, True, 0, 'T', 0, 0.3330000042915344, 0, True, 10, False, 2),
               ('20', 1230237, '.', 'T', '.', 47.0, (False, False, True), 2, False, 0, 'T', 0, 0.0, 0, False, 13, False, 3),
               ('20', 1234567, 'microsat1', 'G', 'GA', 50.0, (False, False, True), 3, False, 1, 'G', 3, 0.0, 6, False, 9, False, 3),
               ('20', 1235237, '.', 'T', '.', 0.0, (False, False, False), 2, False, 0, '.', 0, 0.0, 0, False, 0, False, 0),
               ('X', 10, 'rsTest', 'AC', 'A', 10.0, (False, False, True), 3, False, -1, '.', 0, 0.0, 0, False, 0, False, 0)],
              dtype=[('CHROM', 'S12'), ('POS', '<i4'), ('ID', 'S12'), ('REF', 'S12'), ('ALT', 'S12'), ('QUAL', '<f4'), ('FILTER', [('q10', '?'), ('s50', '?'), ('PASS', '?')]), ('num_alleles', 'u1'), ('is_snp', '?'), ('svlen', '<i4'), ('AA', 'S12'), ('AC', '<u2'), ('AF', '<f4'), ('AN', '<u2'), ('DB', '?'), ('DP', '<i4'), ('H2', '?'), ('NS', '<i4')])
        >>> v['QUAL']
        array([  9.60000038,  10.        ,  29.        ,   3.        ,
                67.        ,  47.        ,  50.        ,   0.        ,  10.        ], dtype=float32)
        >>> v['FILTER']['PASS']
        array([False, False,  True, False,  True,  True,  True, False,  True], dtype=bool)
        >>> v['AF']
        array([ 0.   ,  0.   ,  0.5  ,  0.017,  0.333,  0.   ,  0.   ,  0.   ,  0.   ], dtype=float32)

    """

    loader = _VariantsLoader(vcf_fn, region=region, fields=fields,
                             exclude_fields=exclude_fields, dtypes=dtypes,
                             arities=arities, fills=fills,
                             transformers=transformers, vcf_types=vcf_types,
                             count=count, progress=progress,
                             logstream=logstream, condition=condition,
                             slice_args=slice_args,
                             flatten_filter=flatten_filter, verbose=verbose,
                             cache=cache, cachedir=cachedir,
                             skip_cached=skip_cached)
    return loader.load()


class _ArrayLoader(object):
    """Abstract class providing support for loading an array optionally via a
    cache layer."""

    # to be overridden in subclass
    array_type = None

    def __init__(self, vcf_fn, logstream=None, verbose=True, **kwargs):
        debug('init')
        self.vcf_fn = vcf_fn
        # deal with polymorphic vcf_fn argument
        self.vcf_fns = _filenames_from_arg(vcf_fn)
        self.log = _get_logger(logstream, verbose)
        for k, v in kwargs.items():
            setattr(self, k, v)

    def load(self):
        log = self.log
        array_type = self.array_type
        vcf_fn = self.vcf_fn
        region = self.region
        cache = self.cache
        cachedir = self.cachedir
        skip_cached = self.skip_cached
        if cache:
            log('caching is enabled')
            cache_fn, is_cached = _get_cache(vcf_fn, array_type=array_type,
                                             region=region, cachedir=cachedir,
                                             log=log)
            if not is_cached:
                log('building array')
                arr = self.build()
                log('saving to cache file', cache_fn)
                np.save(cache_fn, arr)
            elif skip_cached:
                log('skipping load from cache file', cache_fn)
                arr = None
            else:
                log('loading from cache file', cache_fn)
                arr = np.load(cache_fn)

        else:
            log('caching is disabled')
            log('building array')
            arr = self.build()

        return arr

    # to be overridden in subclass
    def build(self):
        pass


def _filenames_from_arg(filename):
    """Utility function to deal with polymorphic filenames argument."""
    if isinstance(filename, string_types):
        filenames = [filename]
    elif isinstance(filename, (list, tuple)):
        filenames = filename
    else:
        raise Exception('filename argument must be string, list or tuple')
    for fn in filenames:
        if not os.path.exists(fn):
            raise Exception('file not found: %s' % fn)
    return filenames


# TODO replace this with use of Python standard libary logging support


class _Logger(object):

    def __init__(self, logstream=None):
        if logstream is None:
            logstream = sys.stderr
        self.logstream = logstream

    def __call__(self, *msg):
        s = ('[vcfnp] '
             + str(datetime.now())
             + ' :: '
             + ' '.join([str(m) for m in msg]))
        print(s, file=self.logstream)
        self.logstream.flush()


def _nolog(*args, **kwargs):
    pass


def _get_logger(logstream, verbose):
    if verbose:
        log = _Logger(logstream)
    else:
        log = _nolog
    return log


def _mk_cache_fn(vcf_fn, array_type, region=None, cachedir=None):
    """Utility function to construct a filename for a cache file, given a VCF
    file name (where the original data came from) and other parameters."""
    if cachedir is None:
        # use the VCF file name as the base for a directory name
        cachedir = vcf_fn + config.CACHEDIR_SUFFIX
    if not os.path.exists(cachedir):
        # ensure cache dir exists
        os.makedirs(cachedir)
    else:
        assert os.path.isdir(cachedir), \
            'unexpected error, cache directory is not a directory: %r' \
            % cachedir
    if region is None:
        # loading the whole genome (i.e., all variants)
        cache_fn = os.path.join(cachedir, '%s.npy' % array_type)
    else:
        # loading a specific region
        region = region.replace(':', '_').replace('-', '_')
        cache_fn = os.path.join(cachedir, '%s.%s.npy' % (array_type, region))
    return cache_fn


def _get_cache(vcf_fn, array_type, region, cachedir, log):
    """Utility function to obtain a cache file name and determine whether or
    not a fresh cache file is available."""

    # guard condition
    if isinstance(vcf_fn, (list, tuple)):
        raise Exception(
            'caching only supported when loading from a single VCF file'
        )

    # create cache file name
    cache_fn = _mk_cache_fn(vcf_fn, array_type=array_type, region=region,
                            cachedir=cachedir)

    # decide whether or not a fresh cache file is available
    # (if not, we will parse the VCF and build array from scratch)
    if not os.path.exists(cache_fn):
        log('no cache file found')
        is_cached = False
    elif os.path.getmtime(vcf_fn) > os.path.getmtime(cache_fn):
        is_cached = False
        log('cache file out of date')
    else:
        is_cached = True
        log('cache file available')

    return cache_fn, is_cached


class _VariantsLoader(_ArrayLoader):
    """Class for building variants array."""

    array_type = 'variants'

    def build(self):
        log = self.log

        # open VCF file to inspect header
        vcf_fns = self.vcf_fns
        vcf = PyVariantCallFile(vcf_fns[0])

        # extract FILTER definitions
        filter_ids = vcf.filter_ids
        _warn_duplicates(filter_ids)
        filter_ids = sorted(set(filter_ids))
        if 'PASS' not in filter_ids:
            filter_ids.append('PASS')
        filter_ids = tuple(filter_ids)

        # extract INFO definitions
        _warn_duplicates(vcf.info_ids)
        info_ids = tuple(sorted(set(vcf.info_ids)))
        info_types = vcf.info_types
        info_counts = vcf.info_counts

        # determine which fields to load
        fields = _variants_fields(self.fields, self.exclude_fields, info_ids)

        # determine whether we need to parse the INFO field at all
        parse_info = any([f not in config.STANDARD_VARIANT_FIELDS
                          for f in fields])

        # support for working around VCFs with bad INFO headers
        vcf_types = self.vcf_types
        for f in fields:
            if f not in config.STANDARD_VARIANT_FIELDS and f not in info_ids:
                # fall back to unary string; can be overridden with
                # vcf_types, dtypes and arities args
                info_types[f] = TYPE_STRING
                info_counts[f] = 1
            if vcf_types is not None and f in vcf_types:
                # override type declared in VCF header
                info_types[f] = config.TYPESTRING2KEY[vcf_types[f]]

        # convert to tuples for convenience
        info_types = tuple(info_types[f] if f in info_types else -1
                           for f in fields)
        info_counts = tuple(info_counts[f] if f in info_counts else -1
                            for f in fields)

        # determine expected number of values for each field
        arities = _variants_arities(fields, self.arities, info_counts)

        # determine fill values to use where number of values is less than
        # expectation
        fills = _variants_fills(fields, self.fills, info_types)

        # initialise INFO field transformers
        transformers = _info_transformers(fields, self.transformers)

        # determine dtype to use
        flatten_filter = self.flatten_filter
        dtype = _variants_dtype(fields, self.dtypes, arities, filter_ids,
                                flatten_filter, info_types)

        # set up iterator
        region = self.region
        condition = self.condition
        if condition is not None:
            condition = np.asarray(condition).astype('uint8')
        it = itervariants(vcf_fns, region=region, fields=fields,
                          arities=arities, fills=fills,
                          info_types=info_types, transformers=transformers,
                          filter_ids=filter_ids, flatten_filter=flatten_filter,
                          parse_info=parse_info, condition=condition)

        # slice iterator
        slice_args = self.slice_args
        if slice_args:
            it = islice(it, *slice_args)

        # load array
        arr = _fromiter(it, dtype, self.count, self.progress, log)

        return arr


def _warn_duplicates(fields):
    visited = set()
    for f in fields:
        if f in visited:
            print('WARNING: duplicate definition in header: %s' % f,
                  file=sys.stderr)
        visited.add(f)


def _variants_fields(fields, exclude_fields, info_ids):
    """Utility function to determine which fields to extract when loading
    variants."""
    if fields is None:
        # no fields specified by user
        # by default extract all standard and INFO fields
        fields = config.STANDARD_VARIANT_FIELDS + info_ids
    else:
        # fields have been specified
        for f in fields:
            # check for non-standard fields not declared in INFO header
            if f not in config.STANDARD_VARIANT_FIELDS and f not in info_ids:
                # support extracting INFO even if not declared in header,
                # but warn...
                print('WARNING: no INFO definition found for field %s' % f,
                      file=sys.stderr)
    # process any exclusions
    if exclude_fields is not None:
        fields = [f for f in fields if f not in exclude_fields]
    return tuple(f for f in fields)


def _variants_arities(fields, arities, info_counts):
    """Utility function to determine arities (i.e., number of values to
    expect) for variants fields."""
    if arities is None:
        # no arities specified by user
        arities = dict()
    for f, vcf_count in zip(fields, info_counts):
        if f == 'FILTER':
            arities[f] = 1  # force one value for the FILTER field
        elif f not in arities:
            # arity not specified by user
            if f in config.STANDARD_VARIANT_FIELDS:
                arities[f] = config.DEFAULT_VARIANT_ARITY[f]
            elif vcf_count == NUMBER_ALLELE:
                # default to 1 (biallelic)
                arities[f] = 1
            elif vcf_count <= 0:
                # catch any other cases of non-specific arity
                arities[f] = 1
            else:
                # use arity (i.e., number) specified in INFO header
                arities[f] = vcf_count
    # convert to tuple for zipping with fields
    arities = tuple(arities[f] for f in fields)
    return arities


def _variants_fills(fields, fills, info_types):
    """Utility function to determine fill values for variants fields with
    missing values."""
    if fills is None:
        # no fills specified by user
        fills = dict()
    for f, vcf_type in zip(fields, info_types):
        if f == 'FILTER':
            fills[f] = False
        elif f not in fills:
            if f in config.STANDARD_VARIANT_FIELDS:
                fills[f] = config.DEFAULT_VARIANT_FILL[f]
            else:
                fills[f] = config.DEFAULT_FILL_MAP[vcf_type]
    # convert to tuple for zipping with fields
    fills = tuple(fills[f] for f in fields)
    return fills


def _info_transformers(fields, transformers):
    """Utility function to determine transformer functions for variants
    fields."""
    if transformers is None:
        # no transformers specified by user
        transformers = dict()
    for f in fields:
        if f not in transformers:
            transformers[f] = config.DEFAULT_TRANSFORMER.get(f, None)
    return tuple(transformers[f] for f in fields)


def _variants_dtype(fields, dtypes, arities, filter_ids, flatten_filter,
                    info_types):
    """Utility function to build a numpy dtype for a variants array,
    given user arguments and information available from VCF header."""
    dtype = list()
    for f, n, vcf_type in zip(fields, arities, info_types):
        if f == 'FILTER' and flatten_filter:
            # split FILTER into multiple boolean fields
            for flt in filter_ids:
                nm = 'FILTER_' + flt
                dtype.append((nm, 'b1'))
        elif f == 'FILTER' and not flatten_filter:
            # represent FILTER as a structured field
            t = [(flt, 'b1') for flt in filter_ids]
            dtype.append((f, t))
        else:
            if dtypes is not None and f in dtypes:
                # user overrides default dtype
                t = dtypes[f]
            elif f in config.STANDARD_VARIANT_FIELDS:
                t = config.DEFAULT_VARIANT_DTYPE[f]
            elif f in config.DEFAULT_INFO_DTYPE:
                # known INFO field
                t = config.DEFAULT_INFO_DTYPE[f]
            else:
                t = config.DEFAULT_TYPE_MAP[vcf_type]
            # deal with arity
            if n == 1:
                dtype.append((f, t))
            else:
                dtype.append((f, t, (n,)))
    return dtype


def _fromiter(it, dtype, count, progress, log):
    """Utility function to load an array from an iterator."""
    if progress > 0:
        it = _iter_withprogress(it, progress, log)
    if count is not None:
        a = np.fromiter(it, dtype=dtype, count=count)
    else:
        a = np.fromiter(it, dtype=dtype)
    return a


def _iter_withprogress(iterable, progress, log):
    """Utility function to load an array from an iterator, reporting progress
    as we go."""
    before_all = time.time()
    before = before_all
    n = 0
    for i, o in enumerate(iterable):
        yield o
        n = i+1
        if n % progress == 0:
            after = time.time()
            log('%s rows in %.2fs; batch in %.2fs (%d rows/s)'
                % (n, after-before_all, after-before, progress/(after-before)))
            before = after
    after_all = time.time()
    log('%s rows in %.2fs (%d rows/s)'
        % (n, after_all-before_all, n/(after_all-before_all)))


def calldata(vcf_fn, region=None, samples=None, ploidy=2, fields=None,
             exclude_fields=None, dtypes=None, arities=None, fills=None,
             vcf_types=None, count=None, progress=0, logstream=None,
             condition=None, slice_args=None, verbose=True, cache=False,
             cachedir=None, skip_cached=False):
    """
    Load a numpy 1-dimensional structured array with data from the sample
    columns of a VCF file.

    Parameters
    ----------

    vcf_fn: string or list
        Name of the VCF file or list of file names.
    region: string
        Region to extract, e.g., 'chr1' or 'chr1:0-100000'.
    fields: list or array-like
        List of fields to extract from the VCF.
    exclude_fields: list or array-like
        Fields to exclude from extraction.
    dtypes: dict or dict-like
        Dictionary cotaining dtypes to use instead of the default inferred ones
    arities: dict or dict-like
        Override the amount of values to expect.
    fills: dict or dict-like
        Dictionary containing field:fillvalue mappings used to override the
        default fill in values in VCF fields.
    vcf_types: dict or dict-like
        Dictionary containing field:string mappings used to override any
        bogus type declarations in the VCF header.
    count: int
        Attempt to extract a specific number of records.
    progress: int
        If greater than 0, log parsing progress.
    logstream: file or file-like object
        Stream to use for logging progress.
    condition: array
        Boolean array defining which rows to load.
    slice_args: tuple or list
        Slice of the underlying iterator, e.g., (0, 1000, 10) takes every
        10th row from the first 1000.
    verbose: bool
        Log more messages.
    cache: bool
        If True, save the resulting numpy array to disk, and load from the
        cache if present rather than rebuilding from the VCF.
    cachedir: string
        Manually specify the directory to use to store cache files.
    skip_cached: bool
        If True and cache file is fresh, do not load and return None.

    Examples
    --------

        >>> from vcfnp import calldata, view2d
        >>> C = calldata('fixture/sample.vcf')
        >>> C
        array([ ((True, True, [0, 0], 0, 0, '0|0', [10, 10]), (True, True, [0, 0], 0, 0, '0|0', [10, 10]), (True, False, [0, 1], 0, 0, '0/1', [3, 3])),
               ((True, True, [0, 0], 0, 0, '0|0', [10, 10]), (True, True, [0, 0], 0, 0, '0|0', [10, 10]), (True, False, [0, 1], 0, 0, '0/1', [3, 3])),
               ((True, True, [0, 0], 1, 48, '0|0', [51, 51]), (True, True, [1, 0], 8, 48, '1|0', [51, 51]), (True, False, [1, 1], 5, 43, '1/1', [0, 0])),
               ((True, True, [0, 0], 3, 49, '0|0', [58, 50]), (True, True, [0, 1], 5, 3, '0|1', [65, 3]), (True, False, [0, 0], 3, 41, '0/0', [0, 0])),
               ((True, True, [1, 2], 6, 21, '1|2', [23, 27]), (True, True, [2, 1], 0, 2, '2|1', [18, 2]), (True, False, [2, 2], 4, 35, '2/2', [0, 0])),
               ((True, True, [0, 0], 0, 54, '0|0', [56, 60]), (True, True, [0, 0], 4, 48, '0|0', [51, 51]), (True, False, [0, 0], 2, 61, '0/0', [0, 0])),
               ((True, False, [0, 1], 4, 0, '0/1', [0, 0]), (True, False, [0, 2], 2, 17, '0/2', [0, 0]), (False, False, [-1, -1], 3, 40, './.', [0, 0])),
               ((True, False, [0, 0], 0, 0, '0/0', [0, 0]), (True, True, [0, 0], 0, 0, '0|0', [0, 0]), (False, False, [-1, -1], 0, 0, './.', [0, 0])),
               ((True, False, [0, -1], 0, 0, '0', [0, 0]), (True, False, [0, 1], 0, 0, '0/1', [0, 0]), (True, True, [0, 2], 0, 0, '0|2', [0, 0]))],
              dtype=[('NA00001', [('is_called', '?'), ('is_phased', '?'), ('genotype', 'i1', (2,)), ('DP', '<u2'), ('GQ', 'u1'), ('GT', 'S3'), ('HQ', '<i4', (2,))]), ('NA00002', [('is_called', '?'), ('is_phased', '?'), ('genotype', 'i1', (2,)), ('DP', '<u2'), ('GQ', 'u1'), ('GT', 'S3'), ('HQ', '<i4', (2,))]), ('NA00003', [('is_called', '?'), ('is_phased', '?'), ('genotype', 'i1', (2,)), ('DP', '<u2'), ('GQ', 'u1'), ('GT', 'S3'), ('HQ', '<i4', (2,))])])
        >>> C['NA00001']
        array([(True, True, [0, 0], 0, 0, '0|0', [10, 10]),
               (True, True, [0, 0], 0, 0, '0|0', [10, 10]),
               (True, True, [0, 0], 1, 48, '0|0', [51, 51]),
               (True, True, [0, 0], 3, 49, '0|0', [58, 50]),
               (True, True, [1, 2], 6, 21, '1|2', [23, 27]),
               (True, True, [0, 0], 0, 54, '0|0', [56, 60]),
               (True, False, [0, 1], 4, 0, '0/1', [0, 0]),
               (True, False, [0, 0], 0, 0, '0/0', [0, 0]),
               (True, False, [0, -1], 0, 0, '0', [0, 0])],
              dtype=[('is_called', '?'), ('is_phased', '?'), ('genotype', 'i1', (2,)), ('DP', '<u2'), ('GQ', 'u1'), ('GT', 'S3'), ('HQ', '<i4', (2,))])
        >>> C2d = view2d(C)
        >>> C2d
        array([[(True, True, [0, 0], 0, 0, '0|0', [10, 10]),
                (True, True, [0, 0], 0, 0, '0|0', [10, 10]),
                (True, False, [0, 1], 0, 0, '0/1', [3, 3])],
               [(True, True, [0, 0], 0, 0, '0|0', [10, 10]),
                (True, True, [0, 0], 0, 0, '0|0', [10, 10]),
                (True, False, [0, 1], 0, 0, '0/1', [3, 3])],
               [(True, True, [0, 0], 1, 48, '0|0', [51, 51]),
                (True, True, [1, 0], 8, 48, '1|0', [51, 51]),
                (True, False, [1, 1], 5, 43, '1/1', [0, 0])],
               [(True, True, [0, 0], 3, 49, '0|0', [58, 50]),
                (True, True, [0, 1], 5, 3, '0|1', [65, 3]),
                (True, False, [0, 0], 3, 41, '0/0', [0, 0])],
               [(True, True, [1, 2], 6, 21, '1|2', [23, 27]),
                (True, True, [2, 1], 0, 2, '2|1', [18, 2]),
                (True, False, [2, 2], 4, 35, '2/2', [0, 0])],
               [(True, True, [0, 0], 0, 54, '0|0', [56, 60]),
                (True, True, [0, 0], 4, 48, '0|0', [51, 51]),
                (True, False, [0, 0], 2, 61, '0/0', [0, 0])],
               [(True, False, [0, 1], 4, 0, '0/1', [0, 0]),
                (True, False, [0, 2], 2, 17, '0/2', [0, 0]),
                (False, False, [-1, -1], 3, 40, './.', [0, 0])],
               [(True, False, [0, 0], 0, 0, '0/0', [0, 0]),
                (True, True, [0, 0], 0, 0, '0|0', [0, 0]),
                (False, False, [-1, -1], 0, 0, './.', [0, 0])],
               [(True, False, [0, -1], 0, 0, '0', [0, 0]),
                (True, False, [0, 1], 0, 0, '0/1', [0, 0]),
                (True, True, [0, 2], 0, 0, '0|2', [0, 0])]],
              dtype=[('is_called', '?'), ('is_phased', '?'), ('genotype', 'i1', (2,)), ('DP', '<u2'), ('GQ', 'u1'), ('GT', 'S3'), ('HQ', '<i4', (2,))])
        >>> C2d['genotype']
        array([[[ 0,  0],
                [ 0,  0],
                [ 0,  1]],

               [[ 0,  0],
                [ 0,  0],
                [ 0,  1]],

               [[ 0,  0],
                [ 1,  0],
                [ 1,  1]],

               [[ 0,  0],
                [ 0,  1],
                [ 0,  0]],

               [[ 1,  2],
                [ 2,  1],
                [ 2,  2]],

               [[ 0,  0],
                [ 0,  0],
                [ 0,  0]],

               [[ 0,  1],
                [ 0,  2],
                [-1, -1]],

               [[ 0,  0],
                [ 0,  0],
                [-1, -1]],

               [[ 0, -1],
                [ 0,  1],
                [ 0,  2]]], dtype=int8)
        >>> C2d['genotype'][3, :]
        array([[0, 0],
               [0, 1],
               [0, 0]], dtype=int8)

    """

    loader = _CalldataLoader(vcf_fn, region=region, samples=samples,
                             ploidy=ploidy, fields=fields,
                             exclude_fields=exclude_fields, dtypes=dtypes,
                             arities=arities, fills=fills, vcf_types=vcf_types,
                             count=count, progress=progress,
                             logstream=logstream, condition=condition,
                             slice_args=slice_args, verbose=verbose,
                             cache=cache, cachedir=cachedir,
                             skip_cached=skip_cached)
    arr = loader.load()
    return arr


class _CalldataLoader(_ArrayLoader):

    array_type = 'calldata'

    def build(self):
        log = self.log

        # open VCF file to inspect header
        vcf_fns = self.vcf_fns
        vcf = PyVariantCallFile(vcf_fns[0])

        # extract FORMAT definitions
        _warn_duplicates(vcf.format_ids)
        format_ids = tuple(sorted(set(vcf.format_ids)))
        format_types = vcf.format_types
        format_counts = vcf.format_counts

        # extract sample IDs
        all_samples = vcf.sample_names

        # determine which samples to extract
        samples = self.samples
        if samples is None:
            samples = all_samples
        else:
            # guard against unknown samples requested by user
            for s in samples:
                assert s in all_samples, 'unknown sample: %s' % s
        samples = tuple(samples)
        debug(samples)

        # determine which fields to extract
        fields = _calldata_fields(self.fields, self.exclude_fields, format_ids)

        # support for working around VCFs with bad FORMAT headers
        vcf_types = self.vcf_types
        for f in fields:
            if f not in config.STANDARD_CALLDATA_FIELDS and f not in format_ids:
                # fall back to unary string; can be overridden with
                # vcf_types, dtypes and arities args
                format_types[f] = TYPE_STRING
                format_counts[f] = 1
            if vcf_types is not None and f in vcf_types:
                # override type declared in VCF header
                format_types[f] = config.TYPESTRING2KEY[vcf_types[f]]

        # conveniences
        format_types = tuple(format_types[f] if f in format_types else -1
                             for f in fields)
        format_counts = tuple(format_counts[f] if f in format_counts else -1
                              for f in fields)

        # determine expected number of values for each field
        ploidy = self.ploidy
        arities = _calldata_arities(fields, self.arities, format_counts, ploidy)

        # determine fill values to use where number of values is less than
        # expectation
        fills = _calldata_fills(fields, self.fills, format_types, ploidy)

        # construct a numpy dtype for structured array
        dtype = _calldata_dtype(fields, self.dtypes, format_types, arities,
                                samples, ploidy)

        # set up iterator
        condition = self.condition
        if condition is not None:
            condition = np.asarray(condition).astype('uint8')
        region = self.region
        it = itercalldata(vcf_fns=vcf_fns, region=region, samples=samples,
                          ploidy=ploidy, fields=fields, arities=arities,
                          fills=fills, format_types=format_types,
                          condition=condition)

        # slice iterator
        slice_args = self.slice_args
        if slice_args:
            it = islice(it, *slice_args)

        # build an array from the iterator
        arr = _fromiter(it, dtype, self.count, self.progress, log)

        return arr


def calldata_2d(vcf_fn, region=None, samples=None, ploidy=2, fields=None,
                exclude_fields=None, dtypes=None, arities=None, fills=None,
                vcf_types=None, count=None, progress=0, logstream=None,
                condition=None, slice_args=None, verbose=True, cache=False,
                cachedir=None, skip_cached=False):
    """
    Load a numpy 2-dimensional structured array with data from the sample
    columns of a VCF file. Convenience function, equivalent to calldata()
    followed by view2d(), except that if caching is enabled, files will be
    cached as 2D.

    Parameters
    ----------

    vcf_fn: string or list
        Name of the VCF file or list of file names.
    region: string
        Region to extract, e.g., 'chr1' or 'chr1:0-100000'.
    fields: list or array-like
        List of fields to extract from the VCF.
    exclude_fields: list or array-like
        Fields to exclude from extraction.
    dtypes: dict or dict-like
        Dictionary cotaining dtypes to use instead of the default inferred ones.
    arities: dict or dict-like
        Override the amount of values to expect
    fills: dict or dict-like
        Dictionary containing field:fillvalue mappings used to override the
        default fill in values in VCF fields.
    vcf_types: dict or dict-like
        Dictionary containing field:string mappings used to override any
        bogus type declarations in the VCF header.
    count: int
        Attempt to extract a specific number of records.
    progress: int
        If greater than 0, log parsing progress.
    logstream: file or file-like object
        Stream to use for logging progress.
    condition: array
        Boolean array defining which rows to load.
    slice_args: tuple or list
        Slice of the underlying iterator, e.g., (0, 1000, 10) takes every
        10th row from the first 1000.
    verbose: bool
        Log more messages.
    cache: bool
        If True, save the resulting numpy array to disk, and load from the
        cache if present rather than rebuilding from the VCF.
    cachedir: string
        Manually specify the directory to use to store cache files.
    skip_cached: bool
        If True and cache file is fresh, do not load and return None.

    """

    loader = _Calldata2DLoader(vcf_fn, region=region, samples=samples,
                               ploidy=ploidy, fields=fields,
                               exclude_fields=exclude_fields, dtypes=dtypes,
                               arities=arities, fills=fills,
                               vcf_types=vcf_types, count=count,
                               progress=progress, logstream=logstream,
                               condition=condition, slice_args=slice_args,
                               verbose=verbose, cache=cache, cachedir=cachedir,
                               skip_cached=skip_cached)
    arr = loader.load()
    return arr


class _Calldata2DLoader(_CalldataLoader):

    array_type = 'calldata_2d'

    def build(self):
        arr = super(_Calldata2DLoader, self).build()
        return view2d(arr)


def _calldata_fields(fields, exclude_fields, format_ids):
    """Utility function to determine which calldata (i.e., FORMAT) fields to
    extract."""
    if fields is None:
        # no fields specified by user
        # default to all standard fields plus all FORMAT fields in VCF header
        fields = config.STANDARD_CALLDATA_FIELDS + format_ids
    else:
        # fields specified by user
        for f in fields:
            # check if field is standard or defined in VCF header
            if f not in config.STANDARD_CALLDATA_FIELDS and f not in format_ids:
                # support extracting FORMAT even if not declared in header,
                # but warn...
                print('WARNING: no definition found for field %s' % f,
                      file=sys.stderr)
    # process exclusions
    if exclude_fields is not None:
        fields = [f for f in fields if f not in exclude_fields]
    return tuple(fields)


def _calldata_arities(fields, arities, format_counts, ploidy):
    if arities is None:
        arities = dict()
    for f, vcf_count in zip(fields, format_counts):
        if f not in arities:
            if f == 'genotype':
                arities[f] = ploidy
            elif f in config.DEFAULT_CALLDATA_ARITY:
                arities[f] = config.DEFAULT_CALLDATA_ARITY[f]
            elif vcf_count == NUMBER_ALLELE:
                # default to 2 (biallelic)
                arities[f] = 2
            elif vcf_count == NUMBER_GENOTYPE:
                # arity = (n + p - 1) choose p (n is number of alleles; p is
                # ploidy)
                # default to biallelic (n = 2)
                arities[f] = ploidy + 1
            elif vcf_count <= 0:
                # catch any other cases of non-specific arity
                arities[f] = 1
            else:
                arities[f] = vcf_count
    return tuple(arities[f] for f in fields)


def _calldata_fills(fields, fills, format_types, ploidy):
    if fills is None:
        fills = dict()
    for f, vcf_type in zip(fields, format_types):
        if f not in fills:
            if f == 'GT':
                fills[f] = '/'.join(['.'] * ploidy)
            elif f in config.DEFAULT_CALLDATA_FILL:
                fills[f] = config.DEFAULT_CALLDATA_FILL[f]
            else:
                fills[f] = config.DEFAULT_FILL_MAP[vcf_type]
    return tuple(fills[f] for f in fields)


def _calldata_dtype(fields, dtypes, format_types, arities, samples, ploidy):

    # construct a numpy dtype for structured array cells
    cell_dtype = list()
    for f, vcf_type, n in zip(fields, format_types, arities):
        if dtypes is not None and f in dtypes:
            t = dtypes[f]
        elif f == 'GT':
            t = 'a%d' % ((ploidy*2)-1)
        elif f in config.DEFAULT_CALLDATA_DTYPE:
            # known field
            t = config.DEFAULT_CALLDATA_DTYPE[f]
        else:
            t = config.DEFAULT_TYPE_MAP[vcf_type]
        if n == 1:
            cell_dtype.append((f, t))
        else:
            cell_dtype.append((f, t, (n,)))

    # construct a numpy dtype for structured array
    dtype = [(s, cell_dtype) for s in samples]
    return dtype


def view2d(a):
    """
    Utility function to view a structured 1D array where all fields have a
    uniform dtype (e.g., an array constructed by :func:calldata) as a 2D array.

    Parameters
    ----------

    a: numpy array or array-like
        The array to be viewed as 2D, must have a uniform dtype

    Returns
    -------

    A 2D view of the array.

    Examples
    --------

        >>> from vcfnp import calldata
        >>> a = calldata('sample.vcf')
        >>> a
        array([ ((True, True, [0, 0], '0|0', 0, 0, [10, 10]), (True, True, [0, 0], '0|0', 0, 0, [10, 10]), (True, False, [0, 1], '0/1', 0, 0, [3, 3])),
               ((True, True, [0, 0], '0|0', 0, 0, [10, 10]), (True, True, [0, 0], '0|0', 0, 0, [10, 10]), (True, False, [0, 1], '0/1', 0, 0, [3, 3])),
               ((True, True, [0, 0], '0|0', 48, 1, [51, 51]), (True, True, [1, 0], '1|0', 48, 8, [51, 51]), (True, False, [1, 1], '1/1', 43, 5, [0, 0])),
               ((True, True, [0, 0], '0|0', 49, 3, [58, 50]), (True, True, [0, 1], '0|1', 3, 5, [65, 3]), (True, False, [0, 0], '0/0', 41, 3, [0, 0])),
               ((True, True, [1, 2], '1|2', 21, 6, [23, 27]), (True, True, [2, 1], '2|1', 2, 0, [18, 2]), (True, False, [2, 2], '2/2', 35, 4, [0, 0])),
               ((True, True, [0, 0], '0|0', 54, 0, [56, 60]), (True, True, [0, 0], '0|0', 48, 4, [51, 51]), (True, False, [0, 0], '0/0', 61, 2, [0, 0])),
               ((True, False, [0, 1], '0/1', 0, 4, [0, 0]), (True, False, [0, 2], '0/2', 17, 2, [0, 0]), (True, False, [1, 1], '1/1', 40, 3, [0, 0])),
               ((True, False, [0, 0], '0/0', 0, 0, [0, 0]), (True, True, [0, 0], '0|0', 0, 0, [0, 0]), (False, False, [-1, -1], './.', 0, 0, [0, 0])),
               ((True, False, [0, -1], '0', 0, 0, [0, 0]), (True, False, [0, 1], '0/1', 0, 0, [0, 0]), (True, True, [0, 2], '0|2', 0, 0, [0, 0]))],
              dtype=[('NA00001', [('is_called', '|b1'), ('is_phased', '|b1'), ('genotype', '|i1', (2,)), ('GT', '|S3'), ('GQ', '|u1'), ('DP', '<u2'), ('HQ', '<i4', (2,))]), ('NA00002', [('is_called', '|b1'), ('is_phased', '|b1'), ('genotype', '|i1', (2,)), ('GT', '|S3'), ('GQ', '|u1'), ('DP', '<u2'), ('HQ', '<i4', (2,))]), ('NA00003', [('is_called', '|b1'), ('is_phased', '|b1'), ('genotype', '|i1', (2,)), ('GT', '|S3'), ('GQ', '|u1'), ('DP', '<u2'), ('HQ', '<i4', (2,))])])
        >>> from vcfnp import view2d
        >>> b = view2d(a)
        >>> b
        array([[(True, True, [0, 0], '0|0', 0, 0, [10, 10]),
                (True, True, [0, 0], '0|0', 0, 0, [10, 10]),
                (True, False, [0, 1], '0/1', 0, 0, [3, 3])],
               [(True, True, [0, 0], '0|0', 0, 0, [10, 10]),
                (True, True, [0, 0], '0|0', 0, 0, [10, 10]),
                (True, False, [0, 1], '0/1', 0, 0, [3, 3])],
               [(True, True, [0, 0], '0|0', 48, 1, [51, 51]),
                (True, True, [1, 0], '1|0', 48, 8, [51, 51]),
                (True, False, [1, 1], '1/1', 43, 5, [0, 0])],
               [(True, True, [0, 0], '0|0', 49, 3, [58, 50]),
                (True, True, [0, 1], '0|1', 3, 5, [65, 3]),
                (True, False, [0, 0], '0/0', 41, 3, [0, 0])],
               [(True, True, [1, 2], '1|2', 21, 6, [23, 27]),
                (True, True, [2, 1], '2|1', 2, 0, [18, 2]),
                (True, False, [2, 2], '2/2', 35, 4, [0, 0])],
               [(True, True, [0, 0], '0|0', 54, 0, [56, 60]),
                (True, True, [0, 0], '0|0', 48, 4, [51, 51]),
                (True, False, [0, 0], '0/0', 61, 2, [0, 0])],
               [(True, False, [0, 1], '0/1', 0, 4, [0, 0]),
                (True, False, [0, 2], '0/2', 17, 2, [0, 0]),
                (True, False, [1, 1], '1/1', 40, 3, [0, 0])],
               [(True, False, [0, 0], '0/0', 0, 0, [0, 0]),
                (True, True, [0, 0], '0|0', 0, 0, [0, 0]),
                (False, False, [-1, -1], './.', 0, 0, [0, 0])],
               [(True, False, [0, -1], '0', 0, 0, [0, 0]),
                (True, False, [0, 1], '0/1', 0, 0, [0, 0]),
                (True, True, [0, 2], '0|2', 0, 0, [0, 0])]],
              dtype=[('is_called', '|b1'), ('is_phased', '|b1'), ('genotype', '|i1', (2,)), ('GT', '|S3'), ('GQ', '|u1'), ('DP', '<u2'), ('HQ', '<i4', (2,))])
        >>> b['GT']
        array([['0|0', '0|0', '0/1'],
               ['0|0', '0|0', '0/1'],
               ['0|0', '1|0', '1/1'],
               ['0|0', '0|1', '0/0'],
               ['1|2', '2|1', '2/2'],
               ['0|0', '0|0', '0/0'],
               ['0/1', '0/2', '1/1'],
               ['0/0', '0|0', './.'],
               ['0', '0/1', '0|2']],
              dtype='|S3')

    """

    rows = a.size
    cols = len(a.dtype)
    dtype = a.dtype[0]
    b = a.view(dtype).reshape(rows, cols)
    return b



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division


# standard library dependencies
from itertools import chain
import logging


# internal dependencies
from vcfnp.vcflib import PyVariantCallFile
import vcfnp.config as config
from vcfnp.array import _filenames_from_arg, _warn_duplicates, \
    _variants_fields
from vcfnp.iter import itervariantstable


logger = logging.getLogger(__name__)
debug = logger.debug


class VariantsTable(object):
    def __init__(self, filename, region=None, fields=None, exclude_fields=None,
                 arities=None, flatten_filter=False, fill=b'.',
                 flatteners=None):
        self.filename = filename
        self.region = region
        self.fields = fields
        self.exclude_fields = exclude_fields
        self.arities = arities
        self.flatten_filter = flatten_filter
        self.fill = fill
        self.flatteners = flatteners

    def __iter__(self):

        vcf_fns = _filenames_from_arg(self.filename)

        # extract definitions from VCF header
        vcf = PyVariantCallFile(vcf_fns[0])

        # FILTER definitions
        filter_ids = vcf.filter_ids
        _warn_duplicates(filter_ids)
        filter_ids = sorted(set(filter_ids))
        if 'PASS' not in filter_ids:
            filter_ids.append('PASS')
        filter_ids = tuple(filter_ids)

        # INFO definitions
        _warn_duplicates(vcf.info_ids)
        info_ids = tuple(sorted(set(vcf.info_ids)))
        info_types = vcf.info_types

        # determine fields to extract
        fields = _variants_fields(self.fields, self.exclude_fields, info_ids)

        # turn arities into tuple for convenience
        if self.arities is None:
            arities = (None,) * len(fields)
        else:
            arities = tuple(self.arities.get(f) for f in fields)

        # determine if we need to parse the INFO field
        parse_info = any([f not in config.STANDARD_VARIANT_FIELDS
                          for f in fields])

        # convert to tuple
        info_types = tuple(info_types[f] if f in info_types else -1
                           for f in fields)

        # default flattening
        flatteners = self.flatteners
        if flatteners is None:
            flatteners = dict()
        for f in config.DEFAULT_FLATTEN:
            if f not in flatteners:
                ff, t = config.DEFAULT_FLATTEN[f]
                flatteners[f] = ff, t(self.fill)
        flatteners = tuple(flatteners[f] if f in flatteners else None
                           for f in fields)
        debug(flatteners)

        # make header row
        header = list()
        for f, flattener in zip(fields, flatteners):
            if self.flatten_filter and f == 'FILTER':
                for t in filter_ids:
                    header.append('FILTER_' + t)
            elif (self.arities is not None
                  and f in self.arities
                  and self.arities[f] > 1):
                for i in range(1, self.arities[f] + 1):
                    header.append(f + '_' + str(i))
            elif flattener is not None:
                fflds, _ = flattener
                header.extend(fflds)
            else:
                header.append(f)
        header = tuple(header)
        # make data rows
        data = itervariantstable(vcf_fns=vcf_fns, region=self.region,
                                 fields=fields, arities=arities,
                                 info_types=info_types, parse_info=parse_info,
                                 filter_ids=filter_ids,
                                 flatten_filter=self.flatten_filter,
                                 fill=self.fill, flatteners=flatteners)
        return chain((header,), data)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division


# standard library dependencies
import re
import logging


# internal dependencies
import vcfnp.config as config


logger = logging.getLogger(__name__)
debug = logger.debug


EFF_DEFAULT_DTYPE = [
    ('Effect', 'a33'),
    ('Effect_Impact', 'a8'),
    ('Functional_Class', 'a8'),
    ('Codon_Change', 'a7'),  # N.B., will lose information for indels
    ('Amino_Acid_Change', 'a6'),  # N.B., will lose information for indels
    ('Amino_Acid_Length', 'i4'),
    ('Gene_Name', 'a14'),  # N.B., may be too short for some species
    ('Transcript_BioType', 'a20'),
    ('Gene_Coding', 'i1'),
    ('Transcript_ID', 'a20'),
    ('Exon', 'i1')
]


ANN_DEFAULT_DTYPE = [
    ('Allele', 'a12'),
    ('Annotation', 'a34'),
    ('Annotation_Impact', 'a8'),
    ('Gene_Name', 'a14'),
    ('Gene_ID', 'a14'),
    ('Feature_Type', 'a20'),
    ('Feature_ID', 'a14'),
    ('Transcript_BioType', 'a20'),
    ('Rank', 'i1'),
    ('HGVS_c', 'a12'),
    ('HGVS_p', 'a14'),
    ('cDNA_pos', 'i4'),
    ('cDNA_length', 'i4'),
    ('CDS_pos', 'i4'),
    ('CDS_length', 'i4'),
    ('AA_pos', 'i4'),
    ('AA_length', 'i4'),
    ('Distance', 'i4')
]

config.DEFAULT_INFO_DTYPE['EFF'] = EFF_DEFAULT_DTYPE
config.DEFAULT_VARIANT_ARITY['EFF'] = 1
config.DEFAULT_INFO_DTYPE['ANN'] = ANN_DEFAULT_DTYPE
config.DEFAULT_VARIANT_ARITY['ANN'] = 1


EFF_DEFAULT_FILLS = (b'.', b'.', b'.', b'.', b'.', -1, b'.', b'.', -1, b'.',
                     -1)
ANN_DEFAULT_FILLS = (b'.', b'.', b'.', b'.', b'.', b'.', b'.', b'.', -1, b'.',
                     b'.', -1, -1, -1, -1, -1, -1, -1)


_prog_eff_main = re.compile(b'([^(]+)\(([^)]+)\)')


def eff_default_transformer(fills=EFF_DEFAULT_FILLS):
    """
    Return a simple transformer function for parsing EFF annotations. N.B.,
    ignores all but the first effect.

    """
    def _transformer(vals):
        if len(vals) == 0:
            return fills
        else:
            # ignore all but first effect
            match_eff_main = _prog_eff_main.match(vals[0])
            eff = [match_eff_main.group(1)] \
                  + match_eff_main.group(2).split(b'|')
            result = tuple(
                fill if v == b''
                else int(v) if i == 5 or i == 10
                else (1 if v == b'CODING' else 0) if i == 8
                else v
                for i, (v, fill) in enumerate(list(zip(eff, fills))[:11])
            )
            return result
    return _transformer


def _ann_split2(b):
    x, _, y = b.partition(b'/')
    return [x, y]


def ann_default_transformer(fills=ANN_DEFAULT_FILLS):
    """
    Return a simple transformer function for parsing ANN annotations. N.B.,
    ignores all but the first effect.

    """
    def _transformer(vals):
        if len(vals) == 0:
            return fills
        else:
            # ignore all but first effect
            ann = vals[0].split(b'|')
            ann = ann[:11] + _ann_split2(ann[11]) + _ann_split2(ann[12]) + \
                _ann_split2(ann[13]) + ann[14:]
            result = tuple(
                fill if v == b''
                else int(v.partition(b'/')[0]) if i == 8
                else int(v) if 11 <= i < 18
                else v
                for i, (v, fill) in enumerate(list(zip(ann, fills))[:18])
            )
            return result
    return _transformer


config.DEFAULT_TRANSFORMER['EFF'] = eff_default_transformer()
config.DEFAULT_TRANSFORMER['ANN'] = ann_default_transformer()


EFF_FIELDS = tuple(n for n, _ in EFF_DEFAULT_DTYPE)
# noinspection PyRedeclaration
ANN_FIELDS = tuple(n for n, _ in ANN_DEFAULT_DTYPE)


def flatten_eff(fill=b'.'):
    def _flatten(vals):
        if len(vals) == 0:
            return [fill] * 11
        else:
            # ignore all but first effect
            match_eff_main = _prog_eff_main.match(vals[0])
            eff = [match_eff_main.group(1)] \
                  + match_eff_main.group(2).split(b'|')
            eff = [fill if v == b'' else v for v in eff[:11]]
            return eff
    return _flatten


def flatten_ann(fill=b'.'):
    def _flatten(vals):
        if len(vals) == 0:
            return [fill] * 18
        else:
            # ignore all but first effect
            ann = vals[0].split(b'|')
            ann = ann[:11] + _ann_split2(ann[11]) + _ann_split2(ann[12]) + \
                _ann_split2(ann[13]) + ann[14:]
            ann = [fill if v == b'' else v for v in ann[:18]]
            return ann
    return _flatten


config.DEFAULT_FLATTEN['EFF'] = (EFF_FIELDS, flatten_eff)
config.DEFAULT_FLATTEN['ANN'] = (ANN_FIELDS, flatten_ann)



from cpython.version cimport PY_MAJOR_VERSION
from cpython cimport PyBytes_Check, PyUnicode_Check, PyList_Check, \
    PyTuple_Check, PyDict_Check


if PY_MAJOR_VERSION < 3:
    string_types = basestring,
else:
    string_types = str,


cpdef object s(object x):
    if x is None:
        return None
    elif PY_MAJOR_VERSION < 3:
        return x
    elif PyBytes_Check(x):
        return x.decode('ascii')
    elif PyList_Check(x):
        return [s(i) for i in x]
    elif PyTuple_Check(x):
        return tuple([s(i) for i in x])
    elif PyDict_Check(x):
        return {s(k): s(v) for k, v in x.items()}
    else:
        return x


cpdef object b(object x):
    if x is None:
        return None
    elif PY_MAJOR_VERSION < 3:
        return x
    elif PyBytes_Check(x):
        return x
    elif PyUnicode_Check(x):
        return x.encode('ascii')
    elif PyList_Check(x):
        return [b(i) for i in x]
    elif PyTuple_Check(x):
        return tuple([b(i) for i in x])
    elif PyDict_Check(x):
        return {b(k): b(v) for k, v in x.items()}
    else:
        return x



from __future__ import print_function, division, absolute_import


from vcfnp.vcflib import PyVariantCallFile, TYPE_INTEGER, TYPE_FLOAT, \
    TYPE_STRING, TYPE_BOOL
from nose.tools import eq_


def test_len():
    vcf = PyVariantCallFile('fixture/sample.vcf')
    eq_(9, len(vcf))
    

def test_metadata():
    vcf = PyVariantCallFile('fixture/sample.vcf')
    eq_(['NS', 'AN', 'AC', 'DP', 'AF', 'AA', 'DB', 'H2'], vcf.info_ids)
    eq_(['GT', 'GQ', 'DP', 'HQ'], vcf.format_ids)
    eq_(['s50', 'q10'], vcf.filter_ids)
    eq_(TYPE_INTEGER, vcf.info_types['NS'])
    eq_(TYPE_FLOAT, vcf.info_types['AF'])
    eq_(TYPE_STRING, vcf.info_types['AA'])
    eq_(TYPE_BOOL, vcf.info_types['DB'])
    eq_(TYPE_STRING, vcf.format_types['GT'])
    eq_(TYPE_INTEGER, vcf.format_types['GQ'])
    eq_(1, vcf.info_counts['AA'])
    eq_(2, vcf.format_counts['HQ'])
    eq_(['NA00001', 'NA00002', 'NA00003'], vcf.sample_names)


def test_fixed_fields():
    vcf = PyVariantCallFile('fixture/sample.vcf')
    v = next(iter(vcf))  # first variant
    eq_('19', v.CHROM)
    eq_(111, v.POS)
    eq_('.', v.ID)
    eq_('A', v.REF)
    eq_(['C'], v.ALT)
    eq_(9.6, v.QUAL)
    eq_(['.'], v.FILTER)  # split in C++
    
    
def test_info():
    vcf = PyVariantCallFile('fixture/sample.vcf')
    v = list(iter(vcf))[4]  # fifth variant
    expect = {'AA': ['T'], 'NS': ['2'], 'DP': ['10'], 'AF': ['0.333', '0.667']}
    eq_(expect, v.INFO)


def test_samples():
    vcf = PyVariantCallFile('fixture/sample.vcf')
    v = list(iter(vcf))[2]  # third variant
    expect = {'GT': ['0|0'], 'HQ': ['51', '51'], 'GQ': ['48'], 'DP': ['1']}
    eq_(expect, v.samples['NA00001'])


def test_region():
    vcf = PyVariantCallFile('fixture/sample.vcf.gz')
    vcf.set_region('20')
    eq_(6, len(vcf))
    vcf.set_region('20:1000000-2000000')
    eq_(4, len(vcf))
    vcf.set_region('20', 1000000, 2000000)
    eq_(4, len(vcf))



#
# The MIT License
#
# Copyright (c) 2011 Seoul National University.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Contact: Hyeshik Chang <hyeshik@snu.ac.kr>

import unittest
import random
import gzip
import tabix

EXAMPLEFILE = '../example.gtf.gz'

def load_example_regions(path):
    alldata = []
    for line in gzip.GzipFile(EXAMPLEFILE):
        fields = line.decode('ascii')[:-1].split('\t')
        seqid = fields[0]
        begin = int(fields[3])
        end = int(fields[4])
        alldata.append((seqid, begin, end, fields[:7]))

    return alldata

def does_overlap(A, B, C, D):
    return (A <= D <= B) or (C <= B <= D)

def sample_test_dataset(regions, ntests):
    seqids = [seqid for seqid, _, _, _ in regions]
    lowerbound = max(0, min(begin for _, begin, _, _ in regions) - 1000)
    upperbound = max(end for _, _, end, _ in regions) + 1000

    tests = []
    for i in range(ntests):
        seqid = random.choice(seqids)
        low = random.randrange(lowerbound, upperbound)
        high = random.randrange(low, upperbound)

        # for 1-based both-end inclusive intervals
        matches = [info for seq, begin, end, info in regions
                   if seqid == seq and does_overlap(begin, end, low, high)]

        tests.append((seqid, low, high, matches))

    return tests

def tbresult2excerpt(tbmatches):
    return [fields[:7] for fields in tbmatches]

class TabixTest(unittest.TestCase):
    regions = load_example_regions(EXAMPLEFILE)
    testset = sample_test_dataset(regions, 500)

    def setUp(self):
        self.tb = tabix.Tabix(EXAMPLEFILE)

    def testQuery(self):
        for seqid, low, high, matches in self.testset:
            tbresult = tbresult2excerpt(self.tb.query(seqid, low, high))
            self.assertEqual(tbresult, matches)

    def testQueryS(self):
        for seqid, low, high, matches in self.testset:
            tbresult = tbresult2excerpt(self.tb.querys('%s:%d-%d' %
                                                       (seqid, low, high)))
            self.assertEqual(tbresult, matches)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python

# Author: Heng Li and Aaron Quinlan
# License: MIT/X11

import sys
from ctypes import *
from ctypes.util import find_library
import glob, platform

def load_shared_library(lib, _path='.', ver='*'):
    """Search for and load the tabix library. The
    expectation is that the library is located in
    the current directory (ie. "./")
    """
    # find from the system path
    path = find_library(lib)
    if (path == None): # if fail, search in the custom directory
        s = platform.system()
        if (s == 'Darwin'): suf = ver+'.dylib'
        elif (s == 'Linux'): suf = '.so'+ver
        candidates = glob.glob(_path+'/lib'+lib+suf);
        if (len(candidates) == 1): path = candidates[0]
        else: return None
    cdll.LoadLibrary(path)
    return CDLL(path)

def tabix_init():
    """Initialize and return a tabix reader object
    for subsequent tabix_get() calls.  
    """
    tabix = load_shared_library('tabix')
    if (tabix == None): return None
    tabix.ti_read.restype = c_char_p
    # on Mac OS X 10.6, the following declarations are required.
    tabix.ti_open.restype = c_void_p
    tabix.ti_querys.argtypes = [c_void_p, c_char_p]
    tabix.ti_querys.restype = c_void_p
    tabix.ti_query.argtypes = [c_void_p, c_char_p, c_int, c_int]
    tabix.ti_query.restype = c_void_p
    tabix.ti_read.argtypes = [c_void_p, c_void_p, c_void_p]
    tabix.ti_iter_destroy.argtypes = [c_void_p]
    tabix.ti_close.argtypes = [c_void_p]
    # FIXME: explicit declarations for APIs not used in this script
    return tabix

# OOP interface
class Tabix:
    def __init__(self, fn, fnidx=0):
        self.tabix = tabix_init();
        if (self.tabix == None):
            sys.stderr.write("[Tabix] Please make sure the shared library is compiled and available.\n")
            return
        self.fp = self.tabix.ti_open(fn, fnidx);

    def __del__(self):
        if (self.tabix): self.tabix.ti_close(self.fp)

    def fetch(self, chr, start=-1, end=-1):
        """Generator function that will yield each interval
        within the requested range from the requested file.
        """
        if (self.tabix == None): return
        if (start < 0): iter = self.tabix.ti_querys(self.fp, chr) # chr looks like: "chr2:1,000-2,000" or "chr2"
        else: iter = self.tabix.ti_query(self.fp, chr, start, end) # chr must be a sequence name
        if (iter == None):        
            sys.stderr.write("[Tabix] Malformatted query or wrong sequence name.\n")
            return
        while (1): # iterate
            s = self.tabix.ti_read(self.fp, iter, 0)
            if (s == None): break
            yield s   
        self.tabix.ti_iter_destroy(iter)

# command-line interface
def main():
    if (len(sys.argv) < 3):
        sys.stderr.write("Usage: tabix.py <in.gz> <reg>\n")
        sys.exit(1)
    
    # report the features in the requested interval
    tabix = Tabix(sys.argv[1])
    for line in tabix.fetch(sys.argv[2]):
        print line

if __name__ == '__main__':
    main()



from __future__ import print_function
from distutils.core import setup
from distutils.extension import Extension
import os
from ast import literal_eval

try:
    from Cython.Build import cythonize
except ImportError:
    print('Cython is required.')
    raise

try:
    import numpy as np
except ImportError:
    print('Numpy is required.')
    raise


def get_version(source='vcfnp/__init__.py'):
    with open(source) as f:
        for line in f:
            if line.startswith('__version__'):
                return literal_eval(line.partition('=')[2].lstrip())
    raise ValueError("__version__ not found")


vcflib_dir = os.path.join(os.getcwd(), 'vcflib')
smithwaterman_dir = os.path.join(vcflib_dir, 'smithwaterman')
tabixpp_dir = os.path.join(vcflib_dir, 'tabixpp')
vcflib_sources = ('Variant.cpp', 'ssw.c', 'ssw_cpp.cpp', 'split.cpp')
smithwaterman_sources = ('BandedSmithWaterman.cpp',
                         'SmithWatermanGotoh.cpp',
                         'Repeats.cpp',
                         'disorder.c',
                         'LeftAlign.cpp',
                         'IndelAllele.cpp')
tabixpp_sources = ('bedidx.c', 'bgzf.c', 'index.c', 'knetfile.c', 'kstring.c',
                   'tabix.cpp')


def get_vcflib_sources():
    sources = list()
    sources += [os.path.join(vcflib_dir, s) for s in vcflib_sources]
    sources += [os.path.join(smithwaterman_dir, s)
                for s in smithwaterman_sources]
    sources += [os.path.join(tabixpp_dir, s) for s in tabixpp_sources]
    return sources


compat_extension = Extension(
    'vcfnp.compat',
    language='c++',
    sources=['vcfnp/compat.pyx']
)


vcflib_extension = Extension(
    'vcfnp.vcflib',
    sources=['vcfnp/vcflib.pyx'] + get_vcflib_sources(),
    language='c++',
    include_dirs=[vcflib_dir, smithwaterman_dir, tabixpp_dir, './vcfnp'],
    libraries=['m', 'z'],
    extra_compile_args=['-O3']
)


iter_extension = Extension(
    'vcfnp.iter',
    sources=['vcfnp/iter.pyx'] + get_vcflib_sources(),
    language='c++',
    include_dirs=[vcflib_dir, smithwaterman_dir, tabixpp_dir, './vcfnp'],
    libraries=['m', 'z'],
    extra_compile_args=['-O3']
)


setup(
    name='vcfnp',
    version=get_version(),
    author='Alistair Miles',
    author_email='alimanfoo@googlemail.com',
    url='https://github.com/alimanfoo/vcfnp',
    license='MIT License',
    description='Load numpy arrays from a VCF (variant call file).',
    long_description=open('README.rst').read(),
    classifiers=['Intended Audience :: Developers',
                 'License :: OSI Approved :: MIT License',
                 'Programming Language :: Python',
                 'Topic :: Software Development :: Libraries :: Python Modules'
                 ],
    package_dir={'': '.'},
    packages=['vcfnp', 'vcfnp.test'],
    ext_modules=cythonize([
        compat_extension,
        vcflib_extension,
        iter_extension,
    ]),
    scripts=['scripts/vcf2npy',
             'scripts/qsub_vcf2npy',
             'scripts/vcfnpy2hdf5',
             'scripts/vcf2csv'],
    include_dirs=[np.get_include()]             
)



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division


from vcfnp.vcflib import TYPE_BOOL, TYPE_FLOAT, TYPE_INTEGER, TYPE_STRING, \
    TYPE_UNKNOWN


# default configuration
CACHEDIR_SUFFIX = '.vcfnp_cache'


# these are the standard fields in the variants array
STANDARD_VARIANT_FIELDS = (
    'CHROM', 'POS', 'ID', 'REF', 'ALT', 'QUAL', 'FILTER',
    'num_alleles', 'is_snp', 'svlen'
)


TYPESTRING2KEY = {
    'Float': TYPE_FLOAT,
    'Integer': TYPE_INTEGER,
    'String': TYPE_STRING,
    'Flag': TYPE_BOOL,
}


# default dtypes for the variants array fields
DEFAULT_VARIANT_DTYPE = {
    'CHROM': 'a12',
    'POS': 'i4',
    'ID': 'a12',
    'REF': 'a12',
    'ALT': 'a12',
    'QUAL': 'f4',
    'num_alleles': 'u1',
    'is_snp': 'b1',
    'svlen': 'i4',
}


# default arities for the variants array fields
DEFAULT_VARIANT_ARITY = {
    'CHROM': 1,
    'POS': 1,
    'ID': 1,
    'REF': 1,
    'ALT': 1,  # default assume biallelic (1 alt allele)
    'QUAL': 1,
    'num_alleles': 1,
    'is_snp': 1,
    'svlen': 1,  # default assume biallelic
}


# default fill values for the variants fields if values are missing
DEFAULT_VARIANT_FILL = {
    'CHROM': b'',
    'POS': 0,
    'ID': b'',
    'REF': b'',
    'ALT': b'',
    'QUAL': 0,
    'num_alleles': 0,
    'is_snp': False,
    'svlen': 0,
}


# default mapping from VCF field types to numpy dtypes
DEFAULT_TYPE_MAP = {
    TYPE_FLOAT: 'f4',
    TYPE_INTEGER: 'i4',
    TYPE_STRING: 'a12',
    TYPE_BOOL: 'b1',
    TYPE_UNKNOWN: 'a12'  # leave as string
}


# default mapping from VCF field types to fill values for missing values
DEFAULT_FILL_MAP = {
    TYPE_FLOAT: 0.,
    TYPE_INTEGER: 0,
    TYPE_STRING: b'',
    TYPE_BOOL: False,
    TYPE_UNKNOWN: b''
}


# default dtypes for some known INFO fields where lower precision is
# acceptable in most cases
DEFAULT_INFO_DTYPE = {
    'ABHet': 'f2',
    'ABHom': 'f2',
    'AC': 'u2',
    'AF': 'f2',
    'AN': 'u2',
    'BaseQRankSum': 'f2',
    'ClippingRankSum': 'f2',
    'Dels': 'f2',
    'FS': 'f2',
    'HRun': 'u1',
    'HaplotypeScore': 'f2',
    'InbreedingCoeff': 'f2',
    'VariantType': 'a12',
    'MLEAC': 'u2',
    'MLEAF': 'f2',
    'MQ': 'f2',
    'MQ0Fraction': 'f2',
    'MQRankSum': 'f2',
    'OND': 'f2',
    'QD': 'f2',
    'RPA': 'u2',
    'RU': 'a12',
    'ReadPosRankSum': 'f2',
}


DEFAULT_TRANSFORMER = dict()


STANDARD_CALLDATA_FIELDS = ('is_called', 'is_phased', 'genotype')


DEFAULT_CALLDATA_DTYPE = {
    'is_called': 'b1',
    'is_phased': 'b1',
    'genotype': 'i1',
    # set some lower precision defaults for known FORMAT fields
    'AD': 'u2',
    'DP': 'u2',
    'GQ': 'u1',
    'MLPSAC': 'u1',
    'MLPSAF': 'f2',
    'MQ0': 'u2',
    'PL': 'u2',
}


DEFAULT_CALLDATA_FILL = {
    'is_called': False,
    'is_phased': False,
    'genotype': -1,
}


DEFAULT_CALLDATA_ARITY = {
    'is_called': 1,
    'is_phased': 1,
    # N.B., set genotype arity to ploidy
    'AD': 2,  # default to biallelic
}


DEFAULT_FLATTEN = dict()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, division


__version__ = '2.1.4'


import vcfnp.config as config
import vcfnp.eff as eff
from vcfnp.vcflib import PyVariantCallFile as VariantCallFile
from vcfnp.array import variants, calldata, calldata_2d, view2d
from vcfnp.table import VariantsTable
